from aerospike import predicates as as_predicates
import pytest

PREDICATE_METHDOS = [
    as_predicates.equals,
    as_predicates.contains,
    as_predicates.between,
    as_predicates.range,
    as_predicates.geo_contains_geojson_point,
    as_predicates.geo_contains_point,
    as_predicates.geo_within_geojson_region,
    as_predicates.geo_within_radius,
]


@pytest.mark.parametrize("predicate", PREDICATE_METHDOS)
def test_invalid_predicate_use(predicate):
    with pytest.raises(TypeError):
        predicate()



# -*- coding: utf-8 -*-

import pytest
from .test_base_class import TestBaseClass

try:
    from collections import Counter
except ImportError:
    from counter26 import Counter

import aerospike
from aerospike import exception as e


class TestGetMany:
    @pytest.fixture(autouse=True)
    def setup(self, request, as_connection):
        self.keys = []
        for i in range(5):
            key = ("test", "demo", i)
            rec = {"name": "name%s" % (str(i)), "age": i}
            as_connection.put(key, rec)
            self.keys.append(key)
        key = ("test", "demo", "float_value")
        as_connection.put(key, {"float_value": 4.3})
        self.keys.append(key)

        def teardown():
            """
            Teardown method.
            """
            for i in range(5):
                key = ("test", "demo", i)
                as_connection.remove(key)
            key = ("test", "demo", "float_value")
            as_connection.remove(key)

        request.addfinalizer(teardown)

    def test_pos_get_many_without_policy(self):

        records = self.as_connection.get_many(self.keys)

        assert isinstance(records, list)
        assert len(records) == 6

    def test_pos_get_many_with_proper_parameters(self):
        """
        Proper call to the method
        """
        records = self.as_connection.get_many(self.keys, {"total_timeout": 30})

        assert isinstance(records, list)
        assert len(records) == 6
        assert Counter([x[0][2] for x in records]) == Counter([0, 1, 2, 3, 4, "float_value"])
        assert records[5][2] == {"float_value": 4.3}

    def test_pos_get_many_with_none_policy(self):

        records = self.as_connection.get_many(self.keys, None)

        assert isinstance(records, list)
        assert len(records) == 6
        assert Counter([x[0][2] for x in records]) == Counter([0, 1, 2, 3, 4, "float_value"])
        assert records[5][2] == {"float_value": 4.3}

    def test_pos_get_many_with_non_existent_keys(self):
        """
        Verify that non existent keys show up in the result
        set with no associated records
        """
        self.keys.append(("test", "demo", "non-existent"))

        records = self.as_connection.get_many(self.keys)

        assert isinstance(records, list)
        assert len(records) == 7
        assert Counter([x[0][2] for x in records]) == Counter([0, 1, 2, 3, 4, "non-existent", "float_value"])
        for x in records:
            if x[0][2] == "non-existent":
                assert x[2] is None

    def test_pos_get_many_with_all_non_existent_keys(self):

        keys = [("test", "demo", "gm_non_existing_key")]

        records = self.as_connection.get_many(keys)

        assert len(records) == 1
        assert records == [
            (
                (
                    "test",
                    "demo",
                    "gm_non_existing_key",
                    bytearray(b"\x8da\xd1\x12\x1a\x8f\xa2\xfc*m\xbc\xc7}\xb0\xc8\x13\x80;'\x07"),
                ),
                None,
                None,
            )
        ]

    def test_pos_get_many_with_initkey_as_digest(self):
        """
        Verify that the method may be called with a key tuple
        with no primary key and only a digest
        """
        keys = []
        key = ("test", "demo", None, bytearray("asd;as[d'as;djk;uyfl", "utf-8"))
        rec = {"name": "name1", "age": 1}
        self.as_connection.put(key, rec)

        keys.append(key)

        key = ("test", "demo", None, bytearray("ase;as[d'as;djk;uyfl", "utf-8"))
        rec = {"name": "name2", "age": 2}
        self.as_connection.put(key, rec)

        keys.append(key)

        records = self.as_connection.get_many(keys)

        for key in keys:
            self.as_connection.remove(key)

        assert isinstance(records, list)
        assert len(records) == 2
        i = 0
        for x in records:
            if i:
                assert x[0][3] == bytearray(b"ase;as[d'as;djk;uyfl")
            else:
                assert x[0][3] == bytearray(b"asd;as[d'as;djk;uyfl")
            i = i + 1

    def test_pos_get_many_with_non_existent_keys_in_middle(self):

        self.keys.append(("test", "demo", "some_key"))

        for i in range(15, 20):
            key = ("test", "demo", i)
            rec = {"name": "name%s" % (str(i)), "age": i}
            self.as_connection.put(key, rec)
            self.keys.append(key)

        records = self.as_connection.get_many(self.keys)

        for i in range(15, 20):
            key = ("test", "demo", i)
            self.as_connection.remove(key)

        assert isinstance(records, list)
        assert len(records) == 12
        assert Counter([x[0][2] for x in records]) == Counter(
            [0, 1, 2, 3, 4, "some_key", 15, 16, 17, 18, 19, "float_value"]
        )
        for x in records:
            if x[0][2] == "some_key":
                assert x[2] is None

    def test_get_many_with_bytearray_key(self):
        """
        Make sure that get many can handle a a key with a bytearray pk
        """
        keys = [("test", "demo", bytearray([1, 2, 3]))]
        for key in keys:
            self.as_connection.put(key, {"byte": "array"})

        records = self.as_connection.get_many(keys)
        self.as_connection.remove(keys[0])

        bytearray_key = records[0][0]
        assert len(bytearray_key) == 4

        bytearray_pk = bytearray_key[2]
        assert bytearray_pk == bytearray([1, 2, 3])

    # TODO: incorrect test
    def test_pos_get_many_with_constructor_batch_direct_and_method_arg(self):
        """
        This sets use batch_direct to true in the constructor
        and sets it to false in the policy argument to the function
        """
        config = {"policies": {"use_batch_direct": True}}
        client_batch_direct = TestBaseClass.get_new_connection(add_config=config)

        policy = {"use_batch_direct": False}  # noqa: F841
        records = client_batch_direct.get_many(self.keys)

        assert isinstance(records, list)
        assert len(records) == 6

        client_batch_direct.close()

    def test_pos_get_many_with_use_batch_direct_as_method_arg(self):

        policies = {"use_batch_direct": True}

        records = self.as_connection.get_many(self.keys, policies)

        assert isinstance(records, list)
        assert len(records) == len(self.keys)

    # Negative Tests
    def test_neg_get_many_Invalid_Key_without_primary_key(self):
        """
        Invoke get_many() without primary key
        """
        key = ("test", "set")
        with pytest.raises(e.ParamError):
            key, _, _ = self.as_connection.get(key)

    def test_neg_get_many_with_proper_parameters_without_connection(self):
        config = TestBaseClass.get_connection_config()
        client1 = aerospike.client(config)
        client1.close()

        with pytest.raises(e.ClusterError):
            client1.get_many(self.keys, {"total_timeout": 20})

    def test_neg_prepend_Invalid_Key_without_set_name(self):
        """
        Invoke prepend() without set name
        """
        key = ("test", 1)
        with pytest.raises(e.ParamError):
            key, _, _ = self.as_connection.get(key)

    def test_neg_get_many_with_invalid_key(self):

        with pytest.raises(e.ParamError):
            self.as_connection.get_many("key")

    def test_neg_get_many_with_invalid_timeout(self):

        policies = {"total_timeout": 0.2}
        with pytest.raises(e.ParamError):
            self.as_connection.get_many(self.keys, policies)

    def test_neg_get_many_without_any_parameter(self):

        with pytest.raises(TypeError) as typeError:
            self.as_connection.get_many()

        assert "argument 'keys' (pos 1)" in str(typeError.value)

    def test_neg_get_many_with_none_keys(self):

        with pytest.raises(e.ParamError):
            self.as_connection.get_many(None, {})

    def test_get_many_with_an_invalid_key_in_list_batch_direct(self):

        with pytest.raises(e.ParamError):
            self.as_connection.get_many([("test", "demo", 1), ("test", "demo", 2), None], {"use_batch_direct": True})

    def test_neg_prepend_Invalid_Key_Invalid_ns(self):
        """
        Invoke prepend() invalid namespace
        """
        key = ("test1", "demo", 1)
        with pytest.raises(e.ClientError):
            key, _, _ = self.as_connection.get(key)



# Insert a record with bin1
client.put(keyTuple, {'bin1': 4})

# Insert another bin named bin2
client.put(keyTuple, {'bin2': "value"})

# Remove bin1 from this record
client.put(keyTuple, {'bin2': aerospike.null()})

# Removing the last bin should delete this record
client.put(keyTuple, {'bin1': aerospike.null()})



try:
    import cPickle as pickle
except Exception:
    import pickle


class SomeClass(object):
    pass


pos_data = [
    (("test", "demo", 1), {"age": 1, "name": "name1"}),
    (("test", "demo", 2), {"age": 2, "name": "Mr John", "bmi": 3.55}),
    (("test", "demo", "boolean_key"), {"is_present": True}),
    (("test", "demo", "string"), {"place": "New York", "name": "John"}),
    (("test", "demo", "bb"), {"a": ["aa", 2, "aa", 4, "cc", 3, 2, 1]}),
    (("test", "demo", 1), {"age": 1, "name": "name1"}),
    (("test", "demo", 1), {"is_present": None}),
    (("test", "unknown_set", 1), {"a": {"k": [bytearray("askluy3oijs", "utf-8")]}}),
    # Bytearray
    (("test", "demo", bytearray("asd;as[d'as;d", "utf-8")), {"name": "John"}),
    (("test", "demo", "bytes_key"), {"bytes": bytearray("John", "utf-8")}),
    # List Data
    (("test", "demo", "list_key"), {"names": ["John", "Marlen", "Steve"]}),
    (("test", "demo", "list_key"), {"names": [1, 2, 3, 4, 5]}),
    (("test", "demo", "list_key"), {"names": [1.5, 2.565, 3.676, 4, 5.89]}),
    (("test", "demo", "list_key"), {"names": ["John", "Marlen", 1024]}),
    (("test", "demo", "list_key_unicode"), {"a": ["aa", "bb", 1, "bb", "aa"]}),
    (("test", "demo", "objects"), {"objects": [pickle.dumps(SomeClass()), pickle.dumps(SomeClass())]}),
    # Map Data
    (("test", "demo", "map_key"), {"names": {"name": "John", "age": 24}}),
    (("test", "demo", "map_key_float"), {"double_map": {"1": 3.141, "2": 4.123, "3": 6.285}}),
    (("test", "demo", "map_key_unicode"), {"a": {"aa": "11"}, "b": {"bb": "22"}}),
    #        (('test', 'demo', 1),
    #            {'odict': OrderedDict(sorted({'banana': 3, 'apple': 4, 'pear': 1, 'orange': 2}.items(),
    #                key=lambda t: t[0]))}),
    # Tuple Data
    (("test", "demo", "tuple_key"), {"tuple_seq": tuple("abc")}),
    # Set Data
    (("test", "demo", "set_key"), {"set_data": set([1, 2])}),
    (("test", "demo", "fset_key"), {"fset_data": frozenset(["Frankfurt", "Basel", "Freiburg"])}),
    # Hybrid
    (
        ("test", "demo", "multiple_bins"),
        {
            "i": ["nanslkdl", 1, bytearray("asd;as[d'as;d", "utf-8")],
            "s": {"key": "asd';q;'1';"},
            "b": 1234,
            "l": "!@#@#$QSDAsd;as",
        },
    ),
    (
        ("test", "demo", "list_map_key"),
        {
            "names": ["John", "Marlen", "Steve"],
            "names_and_age": [{"name": "John", "age": 24}, {"name": "Marlen", "age": 25}],
        },
    ),
    (("test", "demo", "map_tuple_key"), {"seq": {"bb": tuple("abc")}}),
]

key_neg = [
    ((None, "demo", 1), -2, "namespace must be a string"),
    ((12.34, "demo", 1), -2, "namespace must be a string"),
    ((35, "demo", 1), -2, "namespace must be a string"),
    (([], "demo", 1), -2, "namespace must be a string"),
    (({}, "demo", 1), -2, "namespace must be a string"),
    (((), "demo", 1), -2, "namespace must be a string"),
    (None, -2, "key is invalid"),
    (["test", "demo", "key_as_list"], -2, "key is invalid"),
    (("test", 123, 1), -2, "set must be a string"),
    (("test", 12.36, 1), -2, "set must be a string"),
    (("test", [], 1), -2, "set must be a string"),
    (("test", {}, 1), -2, "set must be a string"),
    (("test", (), 1), -2, "set must be a string"),
    (("test", "demo", None), -2, "either key or digest is required"),
    (("test", "demo"), -2, "key tuple must be (Namespace, Set, Key) or (Namespace, Set, None, Digest)"),
]



# -*- coding: utf-8 -*-
import pytest
import random
from .test_base_class import TestBaseClass
from aerospike import exception as e

import aerospike


class TestListAppend(object):
    @pytest.fixture(autouse=True)
    def setup(self, request, as_connection):
        keys = []
        for i in range(5):
            key = ("test", "demo", i)
            rec = {"name": "name%s" % (str(i)), "contact_no": [i, i + 1], "city": ["Pune", "Dehli"]}
            as_connection.put(key, rec)
            keys.append(key)

        def teardown():
            """
            Teardown method.
            """
            for key in keys:
                try:
                    as_connection.remove(key)
                except e.RecordNotFound:
                    pass

        request.addfinalizer(teardown)

    @pytest.mark.parametrize(
        "key, field, append_value, expected",
        [
            (
                ("test", "demo", 1),  # append integer value to a list
                "contact_no",
                50000,
                {"contact_no": [1, 2, 50000], "name": "name1", "city": ["Pune", "Dehli"]},
            ),
            (
                ("test", "demo", 1),  # string
                "city",
                "Chennai",
                {"contact_no": [1, 2], "name": "name1", "city": ["Pune", "Dehli", "Chennai"]},
            ),
            (
                ("test", "demo", 1),  # Unicode string
                "city",
                "Mumbai",
                {"contact_no": [1, 2], "city": ["Pune", "Dehli", "Mumbai"], "name": "name1"},
            ),
            (
                ("test", "demo", 2),  # float
                "contact_no",
                85.12,
                {"contact_no": [2, 3, 85.12], "city": ["Pune", "Dehli"], "name": "name2"},
            ),
            (
                ("test", "demo", 3),  # map
                "contact_no",
                {"k1": 29},
                {"contact_no": [3, 4, {"k1": 29}], "city": ["Pune", "Dehli"], "name": "name3"},
            ),
            (
                ("test", "demo", 1),  # bytearray
                "contact_no",
                bytearray("asd;as[d'as;d", "utf-8"),
                {"contact_no": [1, 2, bytearray(b"asd;as[d'as;d")], "city": ["Pune", "Dehli"], "name": "name1"},
            ),
            (
                ("test", "demo", 1),  # boolean
                "contact_no",
                False,
                {"contact_no": [1, 2, 0], "city": ["Pune", "Dehli"], "name": "name1"},
            ),
        ],
    )
    def test_pos_list_append(self, key, field, append_value, expected):
        """
        Invoke list_append() append value to a list
        """
        self.as_connection.list_append(key, field, append_value)

        (key, _, bins) = self.as_connection.get(key)

        assert bins == expected

    def test_pos_list_append_list_with_correct_policy(self):
        """
        Invoke list_append() append list with correct policy options
        """
        key = ("test", "demo", 2)
        policy = {
            "timeout": 1000,
            "retry": aerospike.POLICY_RETRY_ONCE,
            "commit_level": aerospike.POLICY_COMMIT_LEVEL_MASTER,
        }
        self.as_connection.list_append(key, "contact_no", [45, 50, 80], {}, policy)

        (key, _, bins) = self.as_connection.get(key)

        assert bins == {"contact_no": [2, 3, [45, 50, 80]], "city": ["Pune", "Dehli"], "name": "name2"}

    def test_pos_list_append_with_nonexistent_key(self):
        """
        Invoke list_append() with non-existent key
        """
        charSet = "abcdefghijklmnopqrstuvwxyz1234567890"
        minLength = 5
        maxLength = 30
        length = random.randint(minLength, maxLength)
        key = ("test", "demo", "".join(map(lambda unused: random.choice(charSet), range(length))) + ".com")
        status = self.as_connection.list_append(key, "abc", 122)
        assert status == 0

        (key, _, bins) = self.as_connection.get(key)
        self.as_connection.remove(key)

        assert status == 0
        assert bins == {"abc": [122]}

    def test_pos_list_append_with_nonexistent_bin(self):
        """
        Invoke list_append() with non-existent bin
        """
        key = ("test", "demo", 1)
        charSet = "abcdefghijklmnopqrstuvwxyz1234567890"
        minLength = 5
        maxLength = 10
        length = random.randint(minLength, maxLength)
        bin = "".join(map(lambda unused: random.choice(charSet), range(length))) + ".com"
        status = self.as_connection.list_append(key, bin, 585)
        assert status == 0

        (key, _, bins) = self.as_connection.get(key)

        assert status == 0
        assert bins == {"contact_no": [1, 2], "name": "name1", "city": ["Pune", "Dehli"], bin: [585]}

    # Negative Tests
    def test_neg_list_append_with_no_parameters(self):
        """
        Invoke list_append() without any mandatory parameters.
        """
        with pytest.raises(TypeError) as typeError:
            self.as_connection.list_append()
        assert "argument 'key' (pos 1)" in str(typeError.value)

    def test_neg_list_append_with_incorrect_policy(self):
        """
        Invoke list_append() with incorrect policy
        """
        key = ("test", "demo", 1)
        policy = {"timeout": 0.5}
        try:
            self.as_connection.list_append(key, "contact_no", "str", {}, policy)

        except e.ParamError as exception:
            assert exception.code == -2
            assert exception.msg == "timeout is invalid"

    def test_neg_list_append_with_extra_parameter(self):
        """
        Invoke list_append() with extra parameter.
        """
        key = ("test", "demo", 1)
        policy = {"timeout": 1000}
        with pytest.raises(TypeError) as typeError:
            self.as_connection.list_append(key, "contact_no", 999, {}, policy, "")

        assert "list_append() takes at most 5 arguments (6 given)" in str(typeError.value)

    def test_neg_list_append_policy_is_string(self):
        """
        Invoke list_append() with policy is string
        """
        key = ("test", "demo", 1)
        try:
            self.as_connection.list_append(key, "contact_no", 85, {}, "")

        except e.ParamError as exception:
            assert exception.code == -2
            assert exception.msg == "policy must be a dict"

    def test_neg_list_append_key_is_none(self):
        """
        Invoke list_append() with key is none
        """
        try:
            self.as_connection.list_append(None, "contact_no", 45)

        except e.ParamError as exception:
            assert exception.code == -2
            assert exception.msg == "key is invalid"

    def test_neg_list_append_bin_is_none(self):
        """
        Invoke list_append() with bin is none
        """
        key = ("test", "demo", 1)
        try:
            self.as_connection.list_append(key, None, "str")

        except e.ParamError as exception:
            assert exception.code == -2
            assert exception.msg == "Bin name should be of type string"

    def test_neg_list_append_meta_type_integer(self):
        """
        Invoke list_append() with metadata input is of type integer
        """
        key = ("test", "demo", 1)
        try:
            self.as_connection.list_append(key, "contact_no", 85, 888)

        except e.ParamError as exception:
            assert exception.code == -2
            assert exception.msg == "Metadata should be of type dictionary"

    def test_list_append_with_no_connection(self):
        config = TestBaseClass.get_connection_config()
        client = aerospike.client(config)
        client.close()
        k = ("test", "demo", "no_con")
        with pytest.raises(e.ClusterError):
            client.list_append(k, "bob", "item")



# -*- coding: utf-8 -*-
import pytest
from .as_status_codes import AerospikeStatus
from .test_base_class import TestBaseClass
from aerospike import exception as e

import aerospike


class TestScanInfo(object):

    udf_to_load = "bin_lua.lua"

    # connection_with_udf will remove the udf at the end of the tests
    @pytest.fixture(autouse=True)
    def setup(self, request, connection_with_udf):
        """
        Setup method.
        """
        for i in range(15):
            key = ("test", "demo", i)
            rec = {"age": i}
            connection_with_udf.put(key, rec)
        self.job_id = connection_with_udf.scan_apply("test", "demo", "bin_lua", "mytransform", ["age", 2])

        def teardown():
            """
            Teardown method.
            """
            for i in range(15):
                key = ("test", "demo", i)
                connection_with_udf.remove(key)

        request.addfinalizer(teardown)

    def test_job_info_with_no_parameters(self):
        """
        Invoke job_info() without any mandatory parameters.
        """
        with pytest.raises(TypeError) as typeError:
            self.as_connection.job_info()
        assert "argument 'job_id' (pos 1)" in str(typeError.value)

    @pytest.mark.xfail(reason="This test fails if job_info() finishes in < 1ms")
    def test_job_info_with_small_timeout(self, connection_with_udf):
        """
        Invoke job_info() with correct policy and an expected timeout
        """
        policy = {"timeout": 1}

        self.job_id = connection_with_udf.scan_apply("test", "demo", "bin_lua", "mytransform", ["age", 2], block=False)

        with pytest.raises(e.TimeoutError):
            self.as_connection.job_info(self.job_id, aerospike.JOB_SCAN, policy)

    def test_job_info_with_correct_parameters(self):
        """
        Invoke job_info() with correct parameters
        """
        job_info = self.as_connection.job_info(self.job_id, aerospike.JOB_SCAN)
        valid_statuses = (aerospike.JOB_STATUS_COMPLETED, aerospike.JOB_STATUS_INPROGRESS)
        assert job_info["status"] in valid_statuses

        expected_fields = ("status", "progress_pct", "records_read")
        # Make sure that the fields we are expected are in the returned
        # dict
        for field in expected_fields:
            assert field in job_info

    def test_job_info_with_correct_policy(self):
        """
        Invoke job_info() with correct policy
        """
        policy = {"timeout": 1000}
        job_info = self.as_connection.job_info(self.job_id, aerospike.JOB_SCAN, policy)

        valid_statuses = (aerospike.JOB_STATUS_COMPLETED, aerospike.JOB_STATUS_INPROGRESS)
        assert job_info["status"] in valid_statuses

    def test_job_info_with_incorrect_policy(self):
        """
        Invoke job_info() with incorrect policy
        """

        policy = {"timeout": 0.5}
        with pytest.raises(e.ParamError) as err_info:
            self.as_connection.job_info(self.job_id, aerospike.JOB_SCAN, policy)

        assert err_info.value.code == -2
        assert err_info.value.msg == "timeout is invalid"

    def test_job_info_with_scanid_incorrect(self):
        """
        Invoke job_info() with scan id incorrect,
        this should not raise an error
        """
        response = self.as_connection.job_info(self.job_id + 2, aerospike.JOB_SCAN)

        assert response["status"] == aerospike.JOB_STATUS_COMPLETED

    def test_job_info_with_largeid(self):
        """
        Invoke job_info() with a large scan id,
        this should not raise an error
        """
        response = self.as_connection.job_info(13287138843617152748, aerospike.JOB_SCAN)

        assert response["status"] == aerospike.JOB_STATUS_COMPLETED

    def test_job_info_with_scanid_string(self):
        """
        Invoke job_info() with scan id incorrect
        """

        with pytest.raises(TypeError) as typeError:
            self.as_connection.job_info("string")
        assert any(
            [
                "job_info() argument 1 must be int" in str(typeError.value),
                "job_info() argument 1 must be an int" in str(typeError.value),
            ]
        )

    def test_job_info_with_correct_parameters_without_connection(self):
        """
        Invoke job_info() with correct parameters without connection
        """

        config = TestBaseClass.get_connection_config()
        client1 = aerospike.client(config)
        client1.close()

        with pytest.raises(e.ClusterError) as err_info:
            client1.job_info(self.job_id, aerospike.JOB_SCAN)

        assert err_info.value.code == AerospikeStatus.AEROSPIKE_CLUSTER_ERROR

    def test_job_info_with_constant_out_of_valid_values(self):
        """
        Invoke job_info() with the scan module out of the expected range
        """
        with pytest.raises(e.ParamError):
            self.as_connection.job_info(self.job_id, "not query nor scan")

    @pytest.mark.parametrize("module", (None, 1.5, {}, [], 0))
    def test_job_info_with_module_wrong_type(self, module):
        """
        Invoke job_info() with the scan module argument of the wrong type
        """
        with pytest.raises(TypeError):
            self.as_connection.job_info(self.job_id, module)



from aerospike_helpers.operations import operations

# Add name, update age, and return attributes
client.put(keyTuple, {'age': 25, 'career': 'delivery boy'})
ops = [
    operations.increment("age", 1000),
    operations.write("name", "J."),
    operations.prepend("name", "Phillip "),
    operations.append("name", " Fry"),
    operations.read("name"),
    operations.read("career"),
    operations.read("age")
]
(key, meta, bins) = client.operate_ordered(keyTuple, ops)

# Same output for key and meta as operate()
# But read operations are outputted as bin-value pairs
print(bins)
# [('name': 'Phillip J. Fry'), ('career': 'delivery boy'), ('age': 1025)]



# -*- coding: utf-8 -*-

import pytest
from .test_base_class import TestBaseClass
from aerospike import exception as e
from aerospike_helpers.expressions import (
    And,
    Eq,
    GE,
    GT,
    HLLAdd,
    HLLBin,
    HLLDescribe,
    HLLGetCount,
    HLLGetIntersectCount,
    HLLGetSimilarity,
    HLLGetUnion,
    HLLGetUnionCount,
    HLLInit,
    HLLMayContain,
    LE,
)
from aerospike_helpers.operations import hll_operations
from aerospike_helpers.operations import expression_operations as expressions
from math import sqrt, ceil, floor

import aerospike

# Constants
_NUM_RECORDS = 9


def verify_multiple_expression_result(client, test_ns, test_set, expr, op_bin, expected):
    keys = [(test_ns, test_set, i) for i in range(_NUM_RECORDS + 1)]

    # batch get
    res = [rec for rec in client.get_many(keys, policy={"expressions": expr}) if rec[2]]

    assert len(res) == expected


class TestUsrDefinedClass:

    __test__ = False

    def __init__(self, i):
        self.data = i


class TestExpressions(TestBaseClass):
    @pytest.fixture(autouse=True)
    def setup(self, request, as_connection):
        self.test_ns = "test"
        self.test_set = "demo"

        HLL_ops = [
            hll_operations.hll_add("hll_bin", ["key%s" % str(i) for i in range(10000)], 15, 49),
            hll_operations.hll_add("hll_bin2", ["key%s" % str(i) for i in range(5000, 15000)], 15, 49),
            hll_operations.hll_add("hll_bin3", ["key%s" % str(i) for i in range(20000, 30000)], 15, 49),
        ]

        for i in range(_NUM_RECORDS):
            _, _, _ = self.as_connection.operate((self.test_ns, self.test_set, i), HLL_ops)

        def teardown():
            for i in range(_NUM_RECORDS):
                key = (self.test_ns, self.test_set, i)
                as_connection.remove(key)

        request.addfinalizer(teardown)

    def relative_count_error(self, n_index_bits, expected):
        return expected * (1.04 / sqrt(2**n_index_bits)) * 8

    def relative_intersect_error(self, n_index_bits, bin_counts, bin_intersect_count):
        sigma = 1.04 / sqrt(2**n_index_bits)
        rel_err_sum = 0
        for count in bin_counts:
            rel_err_sum += (sigma * count) ** 2
        rel_err_sum += sigma * (bin_intersect_count**2)

        return sqrt(rel_err_sum)

    @pytest.mark.parametrize(
        "policy, listp, index_bc, mh_bc, bin, expected",
        [
            (None, ["key%s" % str(i) for i in range(11000, 16000)], 15, None, "hll_bin", 15000),
            (None, ["key%s" % str(i) for i in range(11000, 16000)], None, None, "hll_bin", 15000),
            (None, ["key%s" % str(i) for i in range(11000, 16000)], 15, 49, "hll_bin", 15000),
            (
                {"flags": aerospike.HLL_WRITE_NO_FAIL},
                ["key%s" % str(i) for i in range(11000, 16000)],
                None,
                None,
                "hll_bin",
                15000,
            ),
        ],
    )
    def test_hll_add_pos(self, policy, listp, index_bc, mh_bc, bin, expected):
        """
        Test the HLLAdd expression.
        """

        upper_lim = ceil(expected + self.relative_count_error(15, expected))
        lower_lim = floor(expected - self.relative_count_error(15, expected))
        expr = And(
            GE(HLLGetCount(HLLAdd(policy, listp, index_bc, mh_bc, bin)), lower_lim),
            LE(HLLGetCount(HLLAdd(policy, listp, index_bc, mh_bc, bin)), upper_lim),
        )

        verify_multiple_expression_result(
            self.as_connection, self.test_ns, self.test_set, expr.compile(), bin, _NUM_RECORDS
        )

    @pytest.mark.parametrize(
        "policy, index_bc, mh_bc, bin, expected",
        [
            (None, 12, None, "hll_bin", {"hll_bin": [12, 49]}),
            (None, None, None, "hll_bin", {"hll_bin": [15, 49]}),
            (None, 8, 20, "hll_bin", {"hll_bin": [8, 20]}),
            (
                {"flags": aerospike.HLL_WRITE_CREATE_ONLY | aerospike.HLL_WRITE_NO_FAIL},
                15,
                49,
                "hll_bin",
                {"hll_bin": [15, 49]},
            ),
        ],
    )
    def test_hll_init_pos(self, policy, index_bc, mh_bc, bin, expected):
        """
        Test the HLLInit expression.
        """

        if self.server_version < [5, 6]:
            pytest.mark.xfail(reason="Servers older than 5.6 do not support 6.0.0 expressions")
            pytest.xfail()

        expr = HLLDescribe(HLLInit(policy, index_bc, mh_bc, bin))

        ops = [expressions.expression_read(bin, expr.compile())]

        _, _, res = self.as_connection.operate((self.test_ns, self.test_set, 0), ops)
        assert res == expected

    @pytest.mark.parametrize(
        "policy, index_bc, mh_bc, bin, expected",
        [
            # OpNotApplicable because read tries to read failed expression
            ({"flags": aerospike.HLL_WRITE_CREATE_ONLY}, 8, 20, "hll_bin", e.OpNotApplicable)
        ],
    )
    def test_hll_init_neg(self, policy, index_bc, mh_bc, bin, expected):
        """
        Test the HLLInit expression expecting failure.
        """

        if self.server_version < [5, 6]:
            pytest.mark.xfail(reason="Servers older than 5.6 do not support 6.0.0 expressions")
            pytest.xfail()

        expr = HLLDescribe(HLLInit(policy, index_bc, mh_bc, bin))

        ops = [expressions.expression_read(bin, expr.compile())]

        with pytest.raises(expected):
            self.as_connection.operate((self.test_ns, self.test_set, 0), ops)

    @pytest.mark.parametrize(
        "bin, expected, hll_bins",
        [
            ("hll_bin", 25000, ["hll_bin", "hll_bin2", "hll_bin3"]),
            ("hll_bin", 20000, ["hll_bin3"]),
        ],
    )
    def test_hll_get_union_pos(self, bin, expected, hll_bins):
        """
        Test the HLLGetUnion expression.
        """

        upper_lim = ceil(expected + self.relative_count_error(10, expected))
        lower_lim = floor(expected - self.relative_count_error(10, expected))
        record = self.as_connection.get(("test", "demo", 0))
        records = [record[2][hll_bin] for hll_bin in hll_bins] if len(hll_bins) > 1 else record[2][hll_bins[0]]
        expr = And(
            GE(HLLGetCount(HLLGetUnion(records, bin)), lower_lim),
            LE(HLLGetCount(HLLGetUnion(records, bin)), upper_lim),
        )

        verify_multiple_expression_result(
            self.as_connection, self.test_ns, self.test_set, expr.compile(), bin, _NUM_RECORDS
        )

    @pytest.mark.parametrize("bin, expected", [("hll_bin", 25000)])
    def test_hll_get_union_count_pos(self, bin, expected):
        """
        Test the HLLGetUnionCount expression.
        """

        upper_lim = ceil(expected + self.relative_count_error(10, expected))
        lower_lim = floor(expected - self.relative_count_error(10, expected))
        record = self.as_connection.get(("test", "demo", 0))
        records = [record[2]["hll_bin"], record[2]["hll_bin2"], record[2]["hll_bin3"]]
        expr = And(
            GT(HLLGetUnionCount(records, bin), lower_lim),
            LE(HLLGetUnionCount(records, bin), upper_lim),
        )

        verify_multiple_expression_result(
            self.as_connection, self.test_ns, self.test_set, expr.compile(), bin, _NUM_RECORDS
        )

    @pytest.mark.parametrize("bin, expected", [("hll_bin", 5000)])
    def test_hll_get_intersect_count_pos(self, bin, expected):
        """
        Test the HLLGetIntersectCount expression.
        """

        upper_lim = ceil(expected + self.relative_intersect_error(10, [10000, 10000], 5000))
        lower_lim = floor(expected - self.relative_intersect_error(10, [10000, 10000], 5000))
        record = self.as_connection.get(("test", "demo", 0))
        records = [record[2]["hll_bin2"]]
        expr = And(
            GE(HLLGetIntersectCount(records, bin), lower_lim),
            LE(HLLGetIntersectCount(records, bin), upper_lim),
        )

        verify_multiple_expression_result(
            self.as_connection, self.test_ns, self.test_set, expr.compile(), bin, _NUM_RECORDS
        )

    @pytest.mark.parametrize("bin, expected", [("hll_bin", 0.33)])
    def test_hll_get_similarity_pos(self, bin, expected):
        """
        Test the HLLGetSimilarity expression.
        """

        record = self.as_connection.get(("test", "demo", 0))
        records = [record[2]["hll_bin2"]]
        expr = And(
            GE(HLLGetSimilarity(records, bin), expected - 0.03),
            LE(HLLGetSimilarity(records, bin), expected + 0.03),
        )

        verify_multiple_expression_result(
            self.as_connection, self.test_ns, self.test_set, expr.compile(), bin, _NUM_RECORDS
        )

    @pytest.mark.parametrize("bin, expected", [("hll_bin", [15, 49])])
    def test_hll_describe_pos(self, bin, expected):
        """
        Test the HLLDescribe expression.
        """

        expr = Eq(HLLDescribe(bin), expected)
        verify_multiple_expression_result(
            self.as_connection, self.test_ns, self.test_set, expr.compile(), bin, _NUM_RECORDS
        )

    @pytest.mark.parametrize("bin", [("hll_bin")])
    def test_hll_may_contain_pos(self, bin):
        """
        Test the HLLMayContain expression.
        """

        expr = Eq(HLLMayContain(["key1", "key2", "key3"], HLLBin(bin)), 1)
        verify_multiple_expression_result(
            self.as_connection, self.test_ns, self.test_set, expr.compile(), bin, _NUM_RECORDS
        )



# -*- coding: utf-8 -*-
import pytest
import time
import pickle
from .as_status_codes import AerospikeStatus
from aerospike import exception as e
from aerospike import predicates as p
from aerospike_helpers import expressions as exp
from .test_base_class import TestBaseClass

import aerospike


def add_indexes_to_client(client):
    try:
        client.index_integer_create("test", "demo", "age", "test_demo_age_idx")
    except e.IndexFoundError:
        pass

    try:
        client.index_integer_create("test", None, "age", "test_null_age_idx")
    except e.IndexFoundError:
        pass


def create_records(client):
    for i in range(1, 10):
        key = ("test", "demo", i)
        rec = {"name": str(i), "age": i, "val": i}
        client.put(key, rec)

    key = ("test", None, "no_set")
    rec = {"name": "no_set_name", "age": 0}
    client.put(key, rec)


def drop_records(client):
    for i in range(1, 10):
        key = ("test", "demo", i)
        try:
            client.remove(key)
        except e.RecordNotFound:
            pass

    try:
        client.remove(("test", None, "no_set"))
    except e.RecordNotFound:
        pass


def add_test_udf(client):
    policy = {}
    client.udf_put("query_apply.lua", 0, policy)


def drop_test_udf(client):
    client.udf_remove("query_apply.lua")


def add_test_parameter_udf(client):
    policy = {}
    client.udf_put("query_apply_parameters.lua", 0, policy)


def drop_test_parameter_udf(client):
    client.udf_remove("query_apply_parameters.lua")


def remove_indexes_from_client(client):
    client.index_remove("test", "test_demo_age_idx")
    client.index_remove("test", "test_null_age_idx")


class TestQueryApply(object):

    # These functions will run once for this test class, and do all of the
    # required setup and teardown
    connection_setup_functions = (add_test_udf, add_test_parameter_udf, add_indexes_to_client, create_records)
    connection_teardown_functions = (drop_test_udf, drop_test_parameter_udf, remove_indexes_from_client, drop_records)
    age_range_pred = p.between("age", 0, 4)  # Predicate for ages between [0,5)
    no_set_key = ("test", None, "no_set")  # Key for item stored in a namespace but not in a set

    @pytest.fixture(autouse=True)
    def setup(self, request, connection_with_config_funcs):
        client = connection_with_config_funcs
        create_records(client)

    def test_query_apply_with_no_parameters(self):
        """
        Invoke query_apply() without any mandatory parameters.
        It should raise a type error as the wrong parameters are passed
        """
        with pytest.raises(TypeError):
            self.as_connection.query_apply()

    def test_query_apply_with_correct_parameters_no_policy(self):
        """
        Invoke query_apply() with correct parameters.
        It should apply the proper UDF, and
        """
        query_id = self.as_connection.query_apply(
            "test", "demo", self.age_range_pred, "query_apply", "mark_as_applied", ["name", 2]
        )

        self._wait_for_query_complete(query_id)

        self._correct_items_have_been_applied()

    def test_query_apply_with_correct_policy(self):
        """
        Invoke query_apply() with correct policy
        """
        policy = {"total_timeout": 0}
        query_id = self.as_connection.query_apply(
            "test", "demo", self.age_range_pred, "query_apply", "mark_as_applied", ["name", 2], policy
        )

        self._wait_for_query_complete(query_id)
        self._correct_items_have_been_applied()

    def test_query_apply_with_new_expressions(self):
        """
        Invoke query_apply() with correct policy and expressions
        """

        expr = exp.Or(exp.Eq(exp.IntBin("age"), 2), exp.Eq(exp.IntBin("val"), 3))

        policy = {"total_timeout": 0, "expressions": expr.compile()}
        query_id = self.as_connection.query_apply(
            "test", "demo", self.age_range_pred, "query_apply", "mark_as_applied", ["name", 2], policy
        )

        self._wait_for_query_complete(query_id)

        recs = []

        for i in range(1, 10):
            key = ("test", "demo", i)
            _, _, bins = self.as_connection.get(key)
            if bins["name"] == "aerospike":
                recs.append(bins)

        assert len(recs) == 2
        for rec in recs:
            assert rec["age"] == 2 or rec["val"] == 3

    def test_query_apply_with_bad_new_expressions(self):
        """
        Invoke query_apply() with incorrect policy and expressions
        """

        expr = exp.Or(exp.Eq(exp.IntBin(5), 2), exp.Eq(exp.IntBin("val"), 3))

        policy = {"total_timeout": 0, "expressions": expr}
        with pytest.raises(e.ParamError):
            self.as_connection.query_apply(
                "test", "demo", self.age_range_pred, "query_apply", "mark_as_applied", ["name", 2], policy
            )

    def test_query_apply_with_set_argument_as_none(self):
        """
        Invoke query_apply() with correct policy,
        Should casuse no changes as the
        """
        policy = {"total_timeout": 0}
        query_id = self.as_connection.query_apply(
            "test", None, self.age_range_pred, "query_apply", "mark_as_applied", ["name", 2], policy
        )

        self._wait_for_query_complete(query_id)
        self._items_without_set_have_been_applied()

    def test_query_apply_with_incorrect_policy(self):
        """
        Invoke query_apply() with incorrect policy
        """
        policy = {"timeout": 0.5}

        with pytest.raises(e.ParamError):
            self.as_connection.query_apply(
                "test", "demo", self.age_range_pred, "query_apply", "mark_as_applied", ["name", 2], policy
            )

    def test_query_apply_with_nonexistent_set(self):
        """
        Invoke query_apply() with incorrect ns and set
        """
        with pytest.raises(e.NamespaceNotFound):
            self.as_connection.query_apply(
                "test1", "demo1", self.age_range_pred, "query_apply", "mark_as_applied", ["name", 2]
            )

    def test_query_apply_with_incorrect_module_name(self):
        """
        Invoke query_apply() with incorrect module name
        Test that passing an incorrect lua_module does
        not call a function
        """
        query_id = self.as_connection.query_apply(
            "test", "demo", self.age_range_pred, "query_apply_incorrect", "mark_as_applied", ["name", 2]
        )

        self._wait_for_query_complete(query_id)
        self._no_items_have_been_applied()

    def test_query_apply_with_incorrect_function_name(self):
        """
        Invoke query_apply() with incorrect function name,
        does not invoke a different function
        """
        query_id = self.as_connection.query_apply(
            "test", "demo", self.age_range_pred, "query_apply", "mytransform_incorrect", ["name", 2]
        )

        self._wait_for_query_complete(query_id)
        self._no_items_have_been_applied()

    def test_query_apply_with_ns_set_none(self):
        """
        Invoke query_apply() with ns and set as None
        """
        with pytest.raises(TypeError):
            self.as_connection.query_apply(
                None, None, self.age_range_pred, "query_apply", "mark_as_applied", ["name", 2]
            )

    def test_query_apply_with_module_argument_value_is_none(self):
        """
        Invoke query_apply() with None module function
        """

        with pytest.raises(e.ParamError):
            self.as_connection.query_apply("test", "demo", self.age_range_pred, None, None, ["name", 2])

    def test_query_apply_with_too_many_arguments(self):
        """
        Invoke query_apply() with extra argument
        """
        policy = {"timeout": 1000}
        with pytest.raises(TypeError):
            self.as_connection.query_apply(
                "test", "demo", self.age_range_pred, "query_apply", "mytransform_incorrect", ["name", 2], policy, ""
            )

    def test_query_apply_with_udf_arguments_as_string(self):
        """
        Invoke query_apply() with arguments as string
        """
        with pytest.raises(e.ParamError):
            self.as_connection.query_apply("test", "demo", self.age_range_pred, "query_apply", "mark_as_applied", "")

    def test_query_apply_with_udf_argument_as_none(self):
        """
        Invoke query_apply() with arguments as None
        """
        with pytest.raises(e.ParamError):
            self.as_connection.query_apply("test", "demo", self.age_range_pred, "query_apply", "mark_as_applied", None)

    def test_query_apply_with_extra_parameter_to_lua_function(self):
        """
        Invoke query_apply() with extra call to lua
        test that passing an extra argument to a udf does
        not cause the function to fail
        """
        query_id = self.as_connection.query_apply(
            "test", "demo", self.age_range_pred, "query_apply", "mark_as_applied", ["name", 2, 3]
        )

        # time.sleep(2)

        self._wait_for_query_complete(query_id)
        self._correct_items_have_been_applied()

    def test_query_apply_with_missing_parameter_to_function(self):
        """
        Invoke query_apply() with a missing argument
        to a lua function does not cause an error
        """
        query_id = self.as_connection.query_apply(
            "test", "demo", self.age_range_pred, "query_apply", "mark_as_applied_three_arg", ["name", 2]
        )

        # time.sleep(2)

        self._wait_for_query_complete(query_id)
        self._correct_items_have_been_applied()

    def test_query_apply_unicode_literal_for_strings(self):
        """
        Invoke query_apply() with unicode udf
        """
        query_id = self.as_connection.query_apply(
            "test", "demo", self.age_range_pred, "query_apply", "mark_as_applied", ["name", 2]
        )

        self._wait_for_query_complete(query_id)
        self._correct_items_have_been_applied()

    def test_query_apply_with_correct_parameters_without_connection(self):
        """
        Invoke query_apply() with correct parameters without connection
        """
        config = TestBaseClass.get_connection_config()
        client1 = aerospike.client(config)
        client1.close()

        with pytest.raises(e.ClusterError) as err_info:
            client1.query_apply("test", "demo", self.age_range_pred, "query_apply", "mark_as_applied", ["name", 2])

        err_code = err_info.value.code

        assert err_code == AerospikeStatus.AEROSPIKE_CLUSTER_ERROR

    @pytest.mark.parametrize(
        "predicate",
        (
            (),  # Raises system error
            None,
            ("age"),
            ("age", 1),
            ("age", 1, 5),
            (1, 1, "bin"),  # start of a valid predicate
        ),
    )
    def test_invalid_predicate_tuple(self, predicate):

        with pytest.raises(e.ParamError) as err_info:
            self.as_connection.query_apply("test", "demo", predicate, "query_apply", "mark_as_applied", ["name", 2])

        err_code = err_info.value.code
        assert err_code == AerospikeStatus.AEROSPIKE_ERR_PARAM

    def test_stream_udf_parameters(self):
        """
        Invoke query.apply() with a stream udf.
        that accepts additional arguments.
        """
        query_results = (
            self.as_connection.query(
                "test",
                "demo",
            )
            .apply("query_apply_parameters", "query_params", [["age", 5]])
            .results()
        )

        query_results.sort()
        assert query_results == [6, 7, 8, 9]

    def test_stream_udf_parameters_with_set(self):
        """
        Invoke query.apply() with a stream udf.
        arguments contain an unsupported set.
        """
        with pytest.raises(e.ClientError) as err_info:
            self.as_connection.query("test", "demo",).apply(
                "query_apply_parameters",
                "query_params",
                [
                    ["job_type", "job_type", 18],
                    ["id", ["john", {"id", "args", "kwargs", "john"}, ["john", {"mary": 39}]]],
                    [],
                ],
            )

        err_text = err_info.value.msg
        assert "udf function argument type must be supported by Aerospike" in err_text
        err_code = err_info.value.code
        assert err_code == AerospikeStatus.AEROSPIKE_ERR_CLIENT

        with pytest.raises(e.ClientError) as err_info_dict:
            self.as_connection.query("test", "demo",).apply(
                "query_apply_parameters",
                "query_params",
                [
                    ["job_type", "job_type", 18],
                    ["id", ["john", ["john", {"mary": 39, "ken": {"lary", "quinton", "julie", "mark"}}]]],
                    [],
                ],
            )

        err_text = err_info_dict.value.msg
        assert "udf function argument type must be supported by Aerospike" in err_text
        err_code = err_info_dict.value.code
        assert err_code == AerospikeStatus.AEROSPIKE_ERR_CLIENT

    def test_stream_udf_parameters_with_tuple(self):
        """
        Invoke query.apply() with a stream udf.
        arguments contain an unsupported tuple.
        """
        with pytest.raises(e.ClientError) as err_info:
            self.as_connection.query("test", "demo",).apply(
                "query_apply_parameters",
                "query_params",
                [["job_type", "job_type", 18], ["id", ["john", ("id", "args"), ["john", {"mary": 39}]]], []],
            )

        err_text = err_info.value.msg
        assert "udf function argument type must be supported by Aerospike" in err_text
        err_code = err_info.value.code
        assert err_code == AerospikeStatus.AEROSPIKE_ERR_CLIENT

    def test_stream_udf_parameters_with_string(self):
        """
        Invoke query.apply() with a stream udf.
        arguments contain a string not wrapped in a list.
        This should cause an exception.
        """
        with pytest.raises(e.ClientError) as err_info:
            self.as_connection.query(
                "test",
                "demo",
            ).apply("query_apply_parameters", "query_params", "age")

        err_text = err_info.value.msg
        assert "udf function arguments must be enclosed in a list" in err_text
        err_code = err_info.value.code
        assert err_code == AerospikeStatus.AEROSPIKE_ERR_CLIENT

    def test_stream_udf_parameters_with_serialized_set(self):
        """
        Invoke query.apply() with a stream udf.
        arguments contain a serialized set.
        """
        query_results = (
            self.as_connection.query(
                "test",
                "demo",
            )
            .apply(
                "query_apply_parameters",
                "query_params",
                [["age", 5], pickle.dumps({"lary", "quinton", "julie", "mark"})],
            )
            .results()
        )

        query_results.sort()
        assert query_results == [6, 7, 8, 9]

    def test_stream_udf_complicated_parameters(self):
        """
        Invoke query.apply() with a stream udf.
        that accepts additional arguments.
        """
        query_results = (
            self.as_connection.query(
                "test",
                "demo",
            )
            .apply(
                "query_apply_parameters",
                "query_params",
                [["age", 2], ["id", ["john", ["hi"]], ["john", {"mary": 39}]], []],
            )
            .results()
        )

        query_results.sort()
        assert query_results == [3, 4, 5, 6, 7, 8, 9]

    def _correct_items_have_been_applied(self):
        for i in range(1, 5):
            key = ("test", "demo", i)
            _, _, bins = self.as_connection.get(key)
            assert bins["name"] == "aerospike"

        for i in range(5, 10):
            key = ("test", "demo", i)
            _, _, bins = self.as_connection.get(key)
            assert bins["name"] != "aerospike"

        _, _, bins = self.as_connection.get(self.no_set_key)
        assert bins["name"] != "aerospike"

    def _items_without_set_have_been_applied(self):
        for i in range(1, 10):
            key = ("test", "demo", i)
            _, _, bins = self.as_connection.get(key)
            if TestBaseClass.major_ver < 6 or (TestBaseClass.major_ver == 6 and TestBaseClass.minor_ver == 0):
                assert bins["name"] != "aerospike"
            else:
                assert bins["name"] == "aerospike" or bins["name"] == str(i)

        _, _, bins = self.as_connection.get(self.no_set_key)
        assert bins["name"] == "aerospike"

    def _wait_for_query_complete(self, query_id):
        while True:
            response = self.as_connection.job_info(query_id, aerospike.JOB_QUERY)
            if response["status"] != aerospike.JOB_STATUS_INPROGRESS:
                return
            time.sleep(0.1)

    def _no_items_have_been_applied(self):

        for i in range(1, 10):
            key = ("test", "demo", i)
            _, _, bins = self.as_connection.get(key)
            if TestBaseClass.major_ver < 6 or (TestBaseClass.major_ver == 6 and TestBaseClass.minor_ver == 0):
                assert bins["name"] != "aerospike"
            else:
                assert bins["name"] == str(i)

        _, _, bins = self.as_connection.get(self.no_set_key)
        if TestBaseClass.major_ver < 6 or (TestBaseClass.major_ver == 6 and TestBaseClass.minor_ver == 0):
            assert bins["name"] != "aerospike"
        else:
            assert bins["name"] == "no_set_name"



# -*- coding: utf-8 -*-

import pytest
from .test_base_class import TestBaseClass

from aerospike import exception as e
from aerospike_helpers import expressions as exp
from .as_status_codes import AerospikeStatus
from aerospike import predicates as p


import aerospike


def add_sindex(client):
    """
    Load the sindex used in the tests
    """
    try:
        client.index_string_create("test", "demo", "s", "string")
    except e.IndexFoundError:
        pass


def remove_sindex(client):
    """
    Remove the sindex created for these tests
    """
    try:
        client.index_remove("test", "string", {})
    except e.IndexNotFound:
        pass


class TestQueryPartition(TestBaseClass):
    def setup_class(cls):
        # Register setup and teardown functions
        cls.connection_setup_functions = [add_sindex]
        cls.connection_teardown_functions = [remove_sindex]

    @pytest.fixture(autouse=True)
    def setup(self, request, connection_with_config_funcs):
        if self.server_version < [6, 0]:
            pytest.mark.xfail(reason="Servers older than 6.0 do not support partition queries.")
            pytest.xfail()
        as_connection = connection_with_config_funcs

        self.test_ns = "test"
        self.test_set = "demo"

        self.partition_1000_count = 0
        self.partition_1001_count = 0
        self.partition_1002_count = 0
        self.partition_1003_count = 0

        as_connection.truncate(self.test_ns, None, 0)

        for i in range(1, 100000):
            put = 0
            key = (self.test_ns, self.test_set, str(i))
            rec_partition = as_connection.get_key_partition_id(self.test_ns, self.test_set, str(i))

            if rec_partition == 1000:
                self.partition_1000_count += 1
                put = 1
            if rec_partition == 1001:
                self.partition_1001_count += 1
                put = 1
            if rec_partition == 1002:
                self.partition_1002_count += 1
                put = 1
            if rec_partition == 1003:
                self.partition_1003_count += 1
                put = 1
            if put:
                rec = {
                    "i": i,
                    "s": "xyz",
                    "l": [2, 4, 8, 16, 32, None, 128, 256],
                    "m": {"partition": rec_partition, "b": 4, "c": 8, "d": 16},
                }
                as_connection.put(key, rec)
        # print(f"{self.partition_1000_count} records are put in partition 1000, \
        #         {self.partition_1001_count} records are put in partition 1001, \
        #         {self.partition_1002_count} records are put in partition 1002, \
        #         {self.partition_1003_count} records are put in partition 1003")

        def teardown():
            for i in range(1, 100000):
                put = 0
                key = ("test", "demo", str(i))
                rec_partition = as_connection.get_key_partition_id(self.test_ns, self.test_set, str(i))

                if rec_partition == 1000:
                    self.partition_1000_count += 1
                    put = 1
                if rec_partition == 1001:
                    self.partition_1001_count += 1
                    put = 1
                if rec_partition == 1002:
                    self.partition_1002_count += 1
                    put = 1
                if rec_partition == 1003:
                    self.partition_1003_count += 1
                    put = 1
                if put:
                    as_connection.remove(key)

        request.addfinalizer(teardown)

    def test_query_partition_with_existent_ns_and_set(self):

        records = []
        partition_filter = {"begin": 1000, "count": 1}
        policy = {"max_retries": 100, "partition_filter": partition_filter}

        def callback(part_id, input_tuple):
            _, _, record = input_tuple
            records.append(record)

        query_obj = self.as_connection.query(self.test_ns, self.test_set)
        query_obj.max_records = 1000
        query_obj.records_per_second = 4000

        query_obj.foreach(callback, policy)

        assert len(records) == self.partition_1000_count

    def test_query_partition_with_where(self):

        records = []
        partition_filter = {"begin": 1000, "count": 1}
        policy = {"max_retries": 100, "partition_filter": partition_filter}

        def callback(part_id, input_tuple):
            _, _, record = input_tuple
            records.append(record)

        query_obj = self.as_connection.query(self.test_ns, self.test_set)
        query_obj.max_records = 1000
        query_obj.records_per_second = 4000
        query_obj.where(p.equals("s", "xyz"))

        query_obj.foreach(callback, policy)

        assert len(records) == self.partition_1000_count

        part_stats = query_obj.get_partitions_status()

        bval = part_stats[1000][4]
        assert bval != 0

    def test_query_partition_with_short_query(self):

        records = []
        partition_filter = {"begin": 1000, "count": 1}
        policy = {"max_retries": 100, "partition_filter": partition_filter, "short_query": True}

        def callback(part_id, input_tuple):
            _, _, record = input_tuple
            records.append(record)

        query_obj = self.as_connection.query(self.test_ns, self.test_set)
        query_obj.max_records = 100
        query_obj.where(p.equals("s", "xyz"))

        query_obj.foreach(callback, policy)

        assert len(records) == self.partition_1000_count and len(records) <= 100

    def test_query_partition_with_filter(self):

        records = []

        partition_filter = {"begin": 1000, "count": 4}

        expr = exp.Eq(
            exp.MapGetByKey(None, aerospike.MAP_RETURN_VALUE, exp.ResultType.INTEGER, "partition", exp.MapBin("m")),
            1002,
        )

        policy = {"max_retries": 100, "expressions": expr.compile(), "partition_filter": partition_filter}

        def callback(part_id, input_tuple):
            _, _, record = input_tuple
            records.append(record)

        query_obj = self.as_connection.query(self.test_ns, self.test_set)
        query_obj.max_records = 1000
        query_obj.records_per_second = 4000

        query_obj.foreach(callback, policy)

        assert len(records) == self.partition_1002_count

    def test_query_partition_with_filter_zero(self):

        records = []

        partition_filter = {"begin": 1000, "count": 1}

        expr = exp.Eq(
            exp.MapGetByKey(None, aerospike.MAP_RETURN_VALUE, exp.ResultType.INTEGER, "partition", exp.MapBin("m")),
            1002,
        )

        policy = {"max_retries": 100, "expressions": expr.compile(), "partition_filter": partition_filter}

        def callback(part_id, input_tuple):
            _, _, record = input_tuple
            records.append(record)

        query_obj = self.as_connection.query(self.test_ns, self.test_set)
        query_obj.max_records = 1000
        query_obj.records_per_second = 4000

        query_obj.foreach(callback, policy)

        assert len(records) == 0

    def test_query_partition_with_existent_ns_and_none_set(self):

        records = []

        def callback(part_id, input_tuple):
            _, _, record = input_tuple
            records.append(record)

        query_obj = self.as_connection.query(self.test_ns, None)

        query_obj.foreach(callback, {"partition_filter": {"begin": 1000, "count": 1}})

        assert len(records) == self.partition_1000_count

    def test_query_partition_with_timeout_policy(self):

        records = []

        def callback(part_id, input_tuple):
            _, _, record = input_tuple
            records.append(record)

        query_obj = self.as_connection.query(self.test_ns, self.test_set)

        query_obj.foreach(callback, {"timeout": 1001, "partition_filter": {"begin": 1000, "count": 1}})

        assert len(records) == self.partition_1000_count

    # NOTE: This could fail if node record counts are small and unbalanced across nodes.
    # @pytest.mark.xfail(reason="Might fail depending on record count and distribution.")
    def test_query_partition_with_max_records_policy(self):

        records = []

        max_records = self.partition_1000_count // 2

        def callback(part_id, input_tuple):
            _, _, record = input_tuple
            records.append(record)

        query_obj = self.as_connection.query(self.test_ns, self.test_set)
        query_obj.max_records = max_records

        query_obj.foreach(callback, {"partition_filter": {"begin": 1000, "count": 1}})
        assert len(records) <= self.partition_1000_count // 2

    # NOTE: This could fail if node record counts are small and unbalanced across nodes.
    # @pytest.mark.xfail(reason="Might fail depending on record count and distribution.")
    def test_query_partition_with_all_records_policy(self):

        records = []

        max_records = (
            self.partition_1000_count
            + self.partition_1001_count
            + self.partition_1002_count
            + self.partition_1003_count
        )

        def callback(part_id, input_tuple):
            _, _, record = input_tuple
            records.append(record)

        query_obj = self.as_connection.query(self.test_ns, self.test_set)
        query_obj.max_records = max_records

        query_obj.foreach(callback, {"partition_filter": {"begin": 1000, "count": 4}})
        assert len(records) <= max_records

    def test_query_partition_with_socket_timeout_policy(self):

        records = []

        def callback(part_id, input_tuple):
            _, _, record = input_tuple
            records.append(record)

        query_obj = self.as_connection.query(self.test_ns, self.test_set)

        query_obj.foreach(callback, {"socket_timeout": 9876, "partition_filter": {"begin": 1000, "count": 1}})

        assert len(records) == self.partition_1000_count

    def test_query_partition_with_callback_returning_false(self):
        """
        Invoke query() with callback function returns false
        """

        records = []

        def callback(part_id, input_tuple):
            _, _, record = input_tuple
            if len(records) == 10:
                return False
            records.append(record)

        query_obj = self.as_connection.query(self.test_ns, self.test_set)

        query_obj.foreach(callback, {"timeout": 1000, "partition_filter": {"begin": 1000, "count": 1}})
        assert len(records) == 10

    def test_query_partition_with_results_method(self):

        ns = "test"
        st = "demo"

        query_obj = self.as_connection.query(ns, st)

        records = query_obj.results({"partition_filter": {"begin": 1001, "count": 1}})
        assert len(records) == self.partition_1001_count

    def test_resume_part_query(self):
        """
        Resume a query using foreach.
        """
        records = 0
        resumed_records = 0

        def callback(part_id, input_tuple):
            nonlocal records
            if records == 5:
                return False
            records += 1

        query_obj = self.as_connection.query(self.test_ns, self.test_set)

        query_obj.foreach(callback, {"partition_filter": {"begin": 1001, "count": 2}})

        assert records == 5

        partition_status = query_obj.get_partitions_status()

        def resume_callback(part_id, input_tuple):
            nonlocal resumed_records
            resumed_records += 1

        query_obj2 = self.as_connection.query(self.test_ns, self.test_set)

        policy = {
            "partition_filter": {"begin": 1001, "count": 2, "partition_status": partition_status},
        }

        query_obj2.foreach(resume_callback, policy)

        assert records + resumed_records == self.partition_1001_count + self.partition_1002_count

    def test_resume_query_results(self):

        """
        Resume a query using results.
        """
        records = 0

        def callback(part_id, input_tuple):
            nonlocal records
            if records == 5:
                return False
            records += 1

        query_obj = self.as_connection.query(self.test_ns, self.test_set)

        query_obj.foreach(callback, {"partition_filter": {"begin": 1001, "count": 1}})

        assert records == 5

        partition_status = query_obj.get_partitions_status()

        query_obj2 = self.as_connection.query(self.test_ns, self.test_set)

        policy = {
            "partition_filter": {"begin": 1001, "count": 1, "partition_status": partition_status},
        }

        results = query_obj2.results(policy)

        assert records + len(results) == self.partition_1001_count

    def test_query_partition_with_non_existent_ns_and_set(self):

        ns = "namespace"
        st = "set"

        records = []
        query_obj = self.as_connection.query(ns, st)

        def callback(part_id, input_tuple):
            _, _, record = input_tuple
            records.append(record)

        with pytest.raises(e.ClientError) as err_info:
            query_obj.foreach(callback, {"partition_filter": {"begin": 1001, "count": 1}})
        err_code = err_info.value.code
        assert err_code == AerospikeStatus.AEROSPIKE_ERR_CLIENT

    def test_query_partition_with_callback_contains_error(self):
        def callback(part_id, input_tuple):
            raise Exception("callback error")

        query_obj = self.as_connection.query(self.test_ns, self.test_set)

        with pytest.raises(e.ClientError) as err_info:
            query_obj.foreach(callback, {"timeout": 1000, "partition_filter": {"begin": 1001, "count": 1}})

        err_code = err_info.value.code
        assert err_code == AerospikeStatus.AEROSPIKE_ERR_CLIENT

    def test_query_partition_with_callback_non_callable(self):
        query_obj = self.as_connection.query(self.test_ns, self.test_set)

        with pytest.raises(e.ClientError) as err_info:
            query_obj.foreach(5, {"partition_filter": {"begin": 1001, "count": 1}})

        err_code = err_info.value.code
        assert err_code == AerospikeStatus.AEROSPIKE_ERR_CLIENT  # TODO this should be an err param

    def test_query_partition_with_callback_wrong_number_of_args(self):
        def callback(input_tuple):
            pass

        query_obj = self.as_connection.query(self.test_ns, self.test_set)

        with pytest.raises(e.ClientError) as err_info:
            query_obj.foreach(callback, {"partition_filter": {"begin": 1001, "count": 1}})

        err_code = err_info.value.code
        assert err_code == AerospikeStatus.AEROSPIKE_ERR_CLIENT

    @pytest.mark.parametrize(
        "begin",
        [
            (-1),
            (4096),
        ],
    )
    def test_query_partition_with_bad_begin(self, begin):
        records = []
        policy = {"timeout": 1000, "partition_filter": {"begin": begin, "count": 1}}
        query_obj = self.as_connection.query(self.test_ns, self.test_set)

        def callback(part_id, input_tuple):
            _, _, record = input_tuple
            records.append(record)

        with pytest.raises(e.ParamError) as err_info:
            query_obj.foreach(callback, policy)
        err_code = err_info.value.code
        assert err_code == AerospikeStatus.AEROSPIKE_ERR_PARAM
        assert "invalid partition_filter policy begin" in err_info.value.msg

    @pytest.mark.parametrize(
        "count",
        [
            (-1),
            (4097),
        ],
    )
    def test_query_partition_with_bad_count(self, count):
        records = []
        policy = {"timeout": 1000, "partition_filter": {"begin": 0, "count": count}}
        query_obj = self.as_connection.query(self.test_ns, self.test_set)

        def callback(part_id, input_tuple):
            _, _, record = input_tuple
            records.append(record)

        with pytest.raises(e.ParamError) as err_info:
            query_obj.foreach(callback, policy)
        err_code = err_info.value.code
        assert err_code == AerospikeStatus.AEROSPIKE_ERR_PARAM
        assert "invalid partition_filter policy count" in err_info.value.msg

    @pytest.mark.parametrize(
        "begin, count",
        [
            (500, 4096),
            (4095, 2),
        ],
    )
    def test_query_partition_with_bad_range(self, begin, count):
        records = []
        policy = {"timeout": 1000, "partition_filter": {"begin": begin, "count": count}}
        query_obj = self.as_connection.query(self.test_ns, self.test_set)

        def callback(part_id, input_tuple):
            _, _, record = input_tuple
            records.append(record)

        with pytest.raises(e.ParamError) as err_info:
            query_obj.foreach(callback, policy)
        err_code = err_info.value.code
        assert err_code == AerospikeStatus.AEROSPIKE_ERR_PARAM
        assert "invalid partition filter range" in err_info.value.msg

    @pytest.mark.parametrize(
        "p_stats, expected, msg",
        [
            (
                {
                    "done": False,
                    "retry": True,
                    1001: (1001, True, False, bytearray(b"\xe9\xe31\x01sS\xedafw\x00W\xcdM\x80\xd0L\xee\\d"), 0),
                    1002: (
                        1002,
                        "bad_init",
                        False,
                        bytearray(b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"),
                        0,
                    ),
                },
                e.ParamError,
                "invalid init for part_id: 1002",
            ),
            (
                {
                    "done": False,
                    "retry": True,
                    1002: (
                        1002,
                        False,
                        "bad_done",
                        bytearray(b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"),
                        0,
                    ),
                },
                e.ParamError,
                "invalid retry for part_id: 1002",
            ),
            (
                {"done": False, "retry": True, 1003: (1003, False, False, "bad_digest", 0)},
                e.ParamError,
                "invalid digest value for part_id: 1003",
            ),
            (
                {
                    "done": False,
                    "retry": True,
                    1004: (
                        1004,
                        False,
                        False,
                        bytearray(b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"),
                        "bad_bval",
                    ),
                },
                e.ParamError,
                "invalid bval for part_id: 1004",
            ),
            (
                {
                    "done": "bad_done",
                    "retry": True,
                    1004: (
                        1004,
                        False,
                        False,
                        bytearray(b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"),
                        0,
                    ),
                },
                e.ParamError,
                "partition_status dict key 'done' must be an int",
            ),
            (
                {
                    "done": False,
                    "retry": "bad_retry",
                    1004: (
                        1004,
                        False,
                        False,
                        bytearray(b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"),
                        0,
                    ),
                },
                e.ParamError,
                "partition_status dict key 'retry' must be an int",
            ),
            (
                {
                    "retry": True,
                    1004: (
                        1004,
                        False,
                        False,
                        bytearray(b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"),
                        0,
                    ),
                },
                e.ParamError,
                "partition_status dict missing key 'done'",
            ),
            (
                {
                    "done": False,
                    1004: (
                        1004,
                        False,
                        False,
                        bytearray(b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"),
                        0,
                    ),
                },
                e.ParamError,
                "partition_status dict missing key 'retry'",
            ),
        ],
    )
    def test_query_partition_with_bad_status(self, p_stats, expected, msg):
        records = []
        policy = {"timeout": 1000, "partition_filter": {"begin": 1000, "count": 5, "partition_status": p_stats}}
        query_obj = self.as_connection.query(self.test_ns, self.test_set)

        def callback(part_id, input_tuple):
            _, _, record = input_tuple
            records.append(record)

        # query_obj.foreach(callback, policy)
        with pytest.raises(expected) as exc:
            query_obj.foreach(callback, policy)

        assert msg in exc.value.msg



# -*- coding: utf-8 -*-
import pytest
import sys

from .as_status_codes import AerospikeStatus
from .test_base_class import TestBaseClass
from aerospike import exception as e

import aerospike


class TestIncrement(object):
    @pytest.fixture(autouse=True)
    def setup(self, request, as_connection):
        for i in range(5):
            key = ("test", "demo", i)
            rec = {"name": "name%s" % (str(i)), "age": i}
            as_connection.put(key, rec)
        key = ("test", "demo", 6)
        rec = {"age": 6.5}
        as_connection.put(key, rec)

        def teardown():
            """
            Teardown method.
            """
            for i in range(5):
                key = ("test", "demo", i)
                as_connection.remove(key)
            key = ("test", "demo", 6)
            as_connection.remove(key)

        request.addfinalizer(teardown)

    def test_increment_with_no_parameters(self):
        """
        Invoke increment() without any mandatory parameters.
        """
        with pytest.raises(TypeError) as typeError:
            self.as_connection.increment()
        assert "argument 'key' (pos 1)" in str(typeError.value)

    def test_increment_with_correct_parameters(self):
        """
        Invoke increment() with correct parameters
        """
        key = ("test", "demo", 1)
        self.as_connection.increment(key, "age", 5)

        (key, _, bins) = self.as_connection.get(key)

        assert bins == {"age": 6, "name": "name1"}

    def test_increment_with_correct_parameters_float_value(self):
        """
        Invoke increment() with correct parameters and a float value
        """
        if TestIncrement.skip_old_server is True:
            pytest.skip("Server does not support increment on float type")
        key = ("test", "demo", 6)
        self.as_connection.increment(key, "age", 6.4)

        (key, _, bins) = self.as_connection.get(key)

        assert bins == {"age": 12.9}

    def test_increment_with_policy_key_send(self):
        """
        Invoke increment() with policy key send
        """
        key = ("test", "demo", 1)
        policy = {
            "total_timeout": 1000,
            "key": aerospike.POLICY_KEY_SEND,
            "retry": aerospike.POLICY_RETRY_ONCE,
            "commit_level": aerospike.POLICY_COMMIT_LEVEL_MASTER,
        }
        self.as_connection.increment(key, "age", 5, {}, policy)

        (key, _, bins) = self.as_connection.get(key)

        assert bins == {"age": 6, "name": "name1"}
        assert key == (
            "test",
            "demo",
            None,
            bytearray(b"\xb7\xf4\xb88\x89\xe2\xdag\xdeh>\x1d\xf6\x91\x9a\x1e\xac\xc4F\xc8"),
        )

    def test_increment_with_policy_key_digest(self):
        """
        Invoke increment() with policy key digest
        """
        key = ("test", "demo", None, bytearray("asd;as[d'as;djk;uyfl", "utf-8"))
        rec = {"name": "name%s" % (str(1)), "age": 1, "nolist": [1, 2, 3]}
        self.as_connection.put(key, rec)

        policy = {"total_timeout": 1000, "key": aerospike.POLICY_KEY_DIGEST, "retry": aerospike.POLICY_RETRY_NONE}
        self.as_connection.increment(key, "age", 5, {}, policy)

        (key, _, bins) = self.as_connection.get(key)

        assert bins == {"age": 6, "name": "name1", "nolist": [1, 2, 3]}
        assert key == ("test", "demo", None, bytearray(b"asd;as[d'as;djk;uyfl"))
        self.as_connection.remove(key)

    def test_increment_with_correct_policy(self):
        """
        Invoke increment() with correct policy
        """
        key = ("test", "demo", 1)
        policy = {"total_timeout": 1000, "key": aerospike.POLICY_KEY_SEND}
        self.as_connection.increment(key, "age", 5, {}, policy)

        (key, _, bins) = self.as_connection.get(key)

        assert bins == {"age": 6, "name": "name1"}

    def test_increment_with_policy_key_gen_EQ_ignore(self):
        """
        Invoke increment() with gen eq positive ignore
        """
        key = ("test", "demo", 1)
        policy = {
            "total_timeout": 1000,
            "key": aerospike.POLICY_KEY_SEND,
            "retry": aerospike.POLICY_RETRY_ONCE,
            "gen": aerospike.POLICY_GEN_IGNORE,
        }

        meta = {"gen": 10, "ttl": 1200}
        self.as_connection.increment(key, "age", 5, meta, policy)

        (key, meta, bins) = self.as_connection.get(key)

        assert bins == {"age": 6, "name": "name1"}
        assert key == (
            "test",
            "demo",
            None,
            bytearray(b"\xb7\xf4\xb88\x89\xe2\xdag\xdeh>\x1d\xf6\x91\x9a\x1e\xac\xc4F\xc8"),
        )

    def test_increment_with_policy_key_gen_EQ_positive(self):
        """
        Invoke increment() with gen eq positive
        """
        key = ("test", "demo", 1)
        policy = {
            "total_timeout": 1000,
            "key": aerospike.POLICY_KEY_SEND,
            "retry": aerospike.POLICY_RETRY_ONCE,
            "gen": aerospike.POLICY_GEN_EQ,
        }
        (key, meta) = self.as_connection.exists(key)

        gen = meta["gen"]
        meta = {"gen": gen, "ttl": 1200}
        self.as_connection.increment(key, "age", 5, meta, policy)

        (key, meta, bins) = self.as_connection.get(key)

        assert bins == {"age": 6, "name": "name1"}
        assert key == (
            "test",
            "demo",
            None,
            bytearray(b"\xb7\xf4\xb88\x89\xe2\xdag\xdeh>\x1d\xf6\x91\x9a\x1e\xac\xc4F\xc8"),
        )

    def test_increment_with_policy_key_gen_EQ_not_equal(self):
        """
        Invoke increment() with policy key EQ not equal
        """
        key = ("test", "demo", 1)
        policy = {
            "total_timeout": 1000,
            "key": aerospike.POLICY_KEY_SEND,
            "retry": aerospike.POLICY_RETRY_ONCE,
            "gen": aerospike.POLICY_GEN_EQ,
        }
        (key, meta) = self.as_connection.exists(key)
        gen = meta["gen"]

        meta = {"gen": gen + 5, "ttl": 1200}

        #  Since the generations are not equal, this should raise an error
        #  And not increment.
        with pytest.raises(e.RecordGenerationError) as err_info:
            self.as_connection.increment(key, "age", 5, meta, policy)

        err_code = err_info.value.code
        assert err_code == AerospikeStatus.AEROSPIKE_ERR_RECORD_GENERATION

        # The age bin should not have been incremented
        _, _, bins = self.as_connection.get(key)

        assert bins["age"] == 1

    def test_increment_with_policy_key_gen_GT_lesser(self):
        """
        Invoke increment() with gen GT lesser
        """
        key = ("test", "demo", 1)
        policy = {
            "total_timeout": 1000,
            "key": aerospike.POLICY_KEY_SEND,
            "retry": aerospike.POLICY_RETRY_ONCE,
            "gen": aerospike.POLICY_GEN_GT,
        }
        (key, meta) = self.as_connection.exists(key)

        gen = meta["gen"]
        meta = {"gen": gen, "ttl": 1200}
        #  since gen is equal to the server version, this should raise an error
        with pytest.raises(e.RecordGenerationError) as err_info:
            self.as_connection.increment(key, "age", 5, meta, policy)

        err_code = err_info.value.code
        assert err_code == AerospikeStatus.AEROSPIKE_ERR_RECORD_GENERATION

        # The age bin should not have been incremented
        _, _, bins = self.as_connection.get(key)

        assert bins["age"] == 1

    def test_increment_with_policy_key_gen_GT_positive(self):
        """
        Invoke increment() with gen GT positive
        """
        key = ("test", "demo", 1)
        policy = {
            "total_timeout": 1000,
            "key": aerospike.POLICY_KEY_SEND,
            "retry": aerospike.POLICY_RETRY_ONCE,
            "gen": aerospike.POLICY_GEN_GT,
        }
        (key, meta) = self.as_connection.exists(key)

        gen = meta["gen"]
        meta = {"gen": gen + 5, "ttl": 1200}
        self.as_connection.increment(key, "age", 5, meta, policy)

        (key, meta, bins) = self.as_connection.get(key)

        assert bins == {"age": 6, "name": "name1"}
        assert key == (
            "test",
            "demo",
            None,
            bytearray(b"\xb7\xf4\xb88\x89\xe2\xdag\xdeh>\x1d\xf6\x91\x9a\x1e\xac\xc4F\xc8"),
        )

    def test_increment_with_incorrect_policy(self):
        """
        Invoke increment() with incorrect policy
        """
        key = ("test", "demo", 1)
        policy = {"total_timeout": 0.5}
        with pytest.raises(e.ParamError):
            self.as_connection.increment(key, "age", 5, {}, policy)

    def test_increment_with_nonexistent_key(self):
        """
        Invoke increment() with non-existent key
        """
        key = ("test", "demo", "non-existentkey")
        self.as_connection.increment(key, "age", 5)

        (key, _, bins) = self.as_connection.get(key)

        assert bins == {"age": 5}

        self.as_connection.remove(key)

    def test_increment_with_nonexistent_bin(self):
        """
        Invoke increment() with non-existent bin
        """
        key = ("test", "demo", 1)
        self.as_connection.increment(key, "age1", 5)

        (key, _, bins) = self.as_connection.get(key)

        assert bins == {"age1": 5, "name": "name1", "age": 1}

    def test_increment_with_extra_parameter(self):
        """
        Invoke increment() with extra parameter.
        """
        key = ("test", "demo", 1)
        policy = {"total_timeout": 1000}
        with pytest.raises(TypeError) as typeError:
            self.as_connection.increment(key, "age", 2, {}, policy, "")

        assert "increment() takes at most 5 arguments (6 given)" in str(typeError.value)

    def test_increment_policy_is_string(self):
        """
        Invoke increment() with policy is string
        """
        key = ("test", "demo", 1)
        with pytest.raises(e.ParamError) as err_info:
            self.as_connection.increment(key, "age", 2, {}, "")

        err_code = err_info.value.code
        assert err_code == AerospikeStatus.AEROSPIKE_ERR_PARAM

    def test_increment_key_is_none(self):
        """
        Invoke increment() with key is none
        """
        with pytest.raises(e.ParamError) as err_info:
            self.as_connection.increment(None, "age", 2)

        err_code = err_info.value.code
        assert err_code == AerospikeStatus.AEROSPIKE_ERR_PARAM

    def test_increment_bin_is_none(self):
        """
        Invoke increment() with bin is none
        """
        key = ("test", "demo", 1)
        with pytest.raises(e.ParamError) as err_info:
            self.as_connection.increment(key, None, 2)

        err_code = err_info.value.code
        assert err_code == AerospikeStatus.AEROSPIKE_ERR_PARAM

    def test_increment_with_unicode_bin(self):
        """
        Invoke increment() with bin is unicode string
        """
        key = ("test", "demo", 1)
        self.as_connection.increment(key, "age", 10)

        (key, _, bins) = self.as_connection.get(key)

        assert bins == {"age": 11, "name": "name1"}

    def test_increment_with_correct_parameters_without_connection(self):
        """
        Invoke increment() with correct parameters without connection
        """
        key = ("test", "demo", 1)
        config = TestBaseClass.get_connection_config()
        client1 = aerospike.client(config)
        client1.close()

        with pytest.raises(e.ClusterError) as err_info:
            client1.increment(key, "age", 5)

        err_code = err_info.value.code

        assert err_code == AerospikeStatus.AEROSPIKE_CLUSTER_ERROR

    @pytest.mark.skip(reason="This raises a system error." + " Something else should be raised")
    def test_increment_with_integer_greaterthan_maxsize(self):
        """
        Invoke increment() with integer greater then(2^63 - 1)
        """
        key = ("test", "demo", 1)
        bins = {"age": 10}
        self.as_connection.put(key, bins)
        try:
            self.as_connection.increment(key, "age", 68786586756785785745)
        # except SystemError:
        #       pass
        except Exception as exception:
            assert exception.code == AerospikeStatus.AEROSPIKE_ERR_PARAM
            assert exception.msg == "integer value exceeds sys.maxsize"

    def test_increment_with_string_value(self):
        """
        Invoke increment() with string value
        """
        with pytest.raises(e.ParamError) as err_info:
            key = ("test", "demo", 1)
            self.as_connection.increment(key, "age", "5")

        err_code = err_info.value.code
        assert err_code == AerospikeStatus.AEROSPIKE_ERR_PARAM

    def test_increment_non_numeric_value(self):
        """
        Try to increment a string by an int
        """
        key = ("test", "demo", 1)
        with pytest.raises(e.BinIncompatibleType) as err_info:
            self.as_connection.increment(key, "name", 1)

        err_code = err_info.value.code
        assert err_code == AerospikeStatus.AEROSPIKE_ERR_BIN_INCOMPATIBLE_TYPE

    def test_increment_float_value_by_int(self):
        """
        Invoke increment() with correct parameters and a float value
        """
        if TestIncrement.skip_old_server is True:
            pytest.skip("Server does not support increment on float type")
        key = ("test", "demo", 6)
        with pytest.raises(e.BinIncompatibleType):
            self.as_connection.increment(key, "age", 1)

    def test_increment_by_float_val_causing_greater_than_max(self):
        if TestIncrement.skip_old_server is True:
            pytest.skip("Server does not support increment on float type")
        key = ("test", "demo", 6)
        big_number = sys.float_info.max

        self.as_connection.increment(key, "age", big_number)
        self.as_connection.increment(key, "age", big_number)

        _, _, record = self.as_connection.get(key)
        assert record["age"] == float("inf")

    def test_increment_float_value_by_infinity(self):
        """
        Invoke increment() with correct parameters and a float value
        """
        if TestIncrement.skip_old_server is True:
            pytest.skip("Server does not support increment on float type")
        key = ("test", "demo", 6)

        self.as_connection.increment(key, "age", float("inf"))

        _, _, record = self.as_connection.get(key)
        assert record["age"] == float("inf")

    def test_increment_float_value_by_negative_infinity(self):
        """
        Invoke increment() with correct parameters and a float value
        """
        if TestIncrement.skip_old_server is True:
            pytest.skip("Server does not support increment on float type")
        key = ("test", "demo", 6)

        self.as_connection.increment(key, "age", -float("inf"))

        _, _, record = self.as_connection.get(key)
        assert record["age"] == -float("inf")

    def test_increment_int_value_by_float(self):
        """
        Invoke increment() with correct parameters and a float value
        """
        if TestIncrement.skip_old_server is True:
            pytest.skip("Server does not support increment on float type")
        key = ("test", "demo", 1)

        with pytest.raises(e.BinIncompatibleType):
            self.as_connection.increment(key, "age", 1.5)

    @pytest.mark.parametrize("add_value", (bytearray("abc", "utf-8"), None, "a"), ids=["bytes", "None", "string"])
    def test_increment_by_non_numeric_value(self, add_value):
        key = ("test", "demo", 1)
        with pytest.raises(e.ParamError) as err_info:
            self.as_connection.increment(key, "age", add_value)

        err_code = err_info.value.code
        assert err_code == AerospikeStatus.AEROSPIKE_ERR_PARAM

    @pytest.mark.skip(reason="Not sure correct behavior is for overflow")
    def test_increment_past_max_size(self):
        key = ("test", "demo", "overflow")
        record = {"age": 2**63 - 1}
        self.as_connection.put(key, record)

        self.as_connection.increment(key, "age", 2)

        _, _, data = self.as_connection.get(key)
        self.as_connection.remove(key)
        assert data["age"] > 2**63 - 1

    @pytest.mark.skip(reason="Not sure what happens with negative overflow")
    def test_increment_beyond_min_size(self):
        key = ("test", "demo", "overflow")
        record = {"age": -(2**63)}
        self.as_connection.put(key, record)

        self.as_connection.increment(key, "age", -2)

        _, _, data = self.as_connection.get(key)
        self.as_connection.remove(key)
        assert data["age"] < -(2**63)

    def test_increment_with_no_inc_value(self):
        key = ("test", "demo", 1)
        with pytest.raises(TypeError):
            self.as_connection.increment(key, "age")



# -*- coding: utf-8 -*-
import pytest

from .as_status_codes import AerospikeStatus
from .test_base_class import TestBaseClass
from aerospike import exception as e

import aerospike


@pytest.mark.usefixtures("connection_with_udf")
class TestGetRegistered(object):
    def setup_class(cls):
        """
        Set the name of the UDF to load
        store the contents of the file for future comparison
        """
        cls.udf_to_load = "bin_lua.lua"
        cls.loaded_udf_name = "bin_lua.lua"
        cls.udf_language = aerospike.UDF_TYPE_LUA
        with open(cls.udf_to_load, "r") as udf_file:
            cls.loaded_udf_content = udf_file.read()

    def test_udf_get_with_correct_paramters_no_policy(self):
        """
        Invoke udf_get() with correct parameters
        """
        udf_contents = self.as_connection.udf_get(self.loaded_udf_name, self.udf_language)

        # Check for udf file contents
        assert udf_contents == self.loaded_udf_content

    def test_udf_get_with_unicode_string_module_name(self):
        """
        Invoke udf_get() with correct parameters and a policy
        and a unicode module name
        """
        module = "bin_lua.lua"
        policy = {"timeout": 5000}

        udf_contents = self.as_connection.udf_get(module, self.udf_language, policy)

        # Check for udf file contents
        assert udf_contents == self.loaded_udf_content

    def test_udf_get_with_correct_policy(self):
        """
        Invoke udf_get() with correct policy
        """
        policy = {"timeout": 5000}
        udf_contents = self.as_connection.udf_get(self.loaded_udf_name, self.udf_language, policy)

        # Check for udf file contents
        assert udf_contents == self.loaded_udf_content

    def test_udf_get_with_no_parameters(self):
        """
        Invoke udf_get() without any mandatory parameters.
        """
        with pytest.raises(TypeError) as typeError:
            self.as_connection.udf_get()

        assert "argument 'module' (pos 1)" in str(typeError.value)

    def test_udf_get_with_incorrect_policy(self):
        """
        Invoke udf_get() with incorrect policy
        """
        invalid_policy = {"timeout": 0.5}

        with pytest.raises(e.ParamError) as err_info:
            self.as_connection.udf_get(self.loaded_udf_name, self.udf_language, invalid_policy)

        assert err_info.value.code == AerospikeStatus.AEROSPIKE_ERR_PARAM

    def test_udf_get_with_nonexistent_module(self):
        """
        Invoke udf_get() with non-existent module
        """
        non_existant_module = "bin_transform_random_fake_name"
        policy = {"timeout": 1000}

        with pytest.raises(e.UDFError) as err_info:
            self.as_connection.udf_get(non_existant_module, self.udf_language, policy)

        assert err_info.value.code == AerospikeStatus.AEROSPIKE_ERR_UDF

    def test_udf_get_with_invalid_language(self):
        """
        Invoke udf_get() with invalid language code
        """
        invalid_language = 85
        policy = {"timeout": 1000}

        with pytest.raises(e.ClientError) as err_info:
            self.as_connection.udf_get(self.loaded_udf_name, invalid_language, policy)

        assert err_info.value.code == AerospikeStatus.AEROSPIKE_ERR_CLIENT

    def test_udf_get_with_extra_parameter(self):
        """
        Invoke udf_get() with extra parameter.
        """
        policy = {"timeout": 1000}

        # Check for status or empty udf contents
        with pytest.raises(TypeError):
            self.as_connection.udf_get(self.loaded_udf_name, self.udf_language, policy, "")

    @pytest.mark.parametrize("policy", ("", (), [], False, 1))
    def test_udf_get_policy_is_wrong_type(self, policy):
        """
        Invoke udf_get() with policy is string
        """
        with pytest.raises(e.ParamError) as err_info:
            self.as_connection.udf_get(self.loaded_udf_name, self.udf_language, policy)

            assert err_info.value.code == AerospikeStatus.AEROSPIKE_ERR_PARAM

    def test_udf_get_module_is_none(self):
        """
        Invoke udf_get() with module is none
        """
        with pytest.raises(e.ClientError) as err_info:
            self.as_connection.udf_get(None, self.udf_language)

        assert err_info.value.code == AerospikeStatus.AEROSPIKE_ERR_CLIENT

    @pytest.mark.parametrize("udf_module", ([], {}, (), 1, None, ("module",), bytearray("bin_lua.lua", "utf-8")))
    def test_invalid_module_arg_types(self, udf_module):
        with pytest.raises(e.ClientError):
            self.as_connection.udf_get(udf_module, 0)

    @pytest.mark.parametrize("ltype", ("0", None, (), [], {}))
    def test_invalid_language_arg_types(self, ltype):
        with pytest.raises(TypeError):
            self.as_connection.udf_get(self.loaded_udf_name, ltype)

    def test_udf_get_with_correct_paramters_without_connection(self):
        """
        Invoke udf_get() with correct parameters without connection
        """
        policy = {"timeout": 5000}

        config = TestBaseClass.get_connection_config()
        client1 = aerospike.client(config)
        client1.close()

        with pytest.raises(e.ClusterError) as err_info:
            client1.udf_get(self.loaded_udf_name, self.udf_language, policy)

        assert err_info.value.code == AerospikeStatus.AEROSPIKE_CLUSTER_ERROR



# Insert record and get its metadata
client.put(keyTuple, bins = {"bin1": 4})
(key, meta) = client.exists(keyTuple)
print(meta) # {'ttl': 2592000, 'gen': 1}

# Explicitly set TTL to 120
# and increment generation
client.touch(keyTuple, 120)

# Record metadata should be updated
(key, meta) = client.exists(keyTuple)
print(meta) # {'ttl': 120, 'gen': 2}



# -*- coding: utf-8 -*-
import pytest
import aerospike
from aerospike import exception as e
from aerospike_helpers.operations import list_operations
from aerospike_helpers.operations import map_operations as map_ops


def get_list_result_from_operation(client, key, operation, binname):
    _, _, result_bins = client.operate(key, [operation])
    return result_bins[binname]


def skip_less_than_430(version):
    if version < [4, 3]:
        print(version)
        pytest.skip("Requires server > 4.3.0 to work")


class TestNewRelativeListOperations(object):
    @pytest.fixture(autouse=True)
    def setup(self, request, as_connection):
        """
        Setup Method
        """
        skip_less_than_430(self.server_version)
        self.keys = []
        # INDEXES    0, 1, 2, 3, 4, 05
        # RINDEX     5, 4, 3, 2, 1, 0
        # RANK       2, 1, 0, 3, 4, 5
        # RRANK     -4,-5,-6,-3,-2,-1
        self.test_list = [3, 1, 5, 7, 9, 11]

        self.test_key = "test", "demo", "new_list_op"
        self.test_bin = "list"
        self.as_connection.put(self.test_key, {self.test_bin: self.test_list})

        # Make sure the list is ordered, in order to get expected return order.
        ops = [list_operations.list_sort("list", 0), list_operations.list_set_order("list", aerospike.LIST_ORDERED)]
        self.as_connection.operate(self.test_key, ops)
        self.keys.append(self.test_key)

        yield

        for key in self.keys:
            try:
                self.as_connection.remove(key)
            except e.AerospikeError:
                pass

    @pytest.mark.parametrize(
        "value, offset, count, expected",
        (
            (5, 0, None, [5, 7, 9, 11]),
            (5, 0, 2, [5, 7]),
            (4, -2, 3, [1, 3, 5]),
            (0, 0, 3, [1, 3, 5]),
            (150, 0, None, []),
        ),
    )
    def test_list_get_by_value_rank_range_relative(self, value, offset, count, expected):
        """
        Without a return type this should return the value
        """
        operation = list_operations.list_get_by_value_rank_range_relative(
            self.test_bin, value, offset, return_type=aerospike.LIST_RETURN_VALUE, count=count
        )

        result = get_list_result_from_operation(self.as_connection, self.test_key, operation, self.test_bin)

        assert result == expected

    @pytest.mark.parametrize(
        "value, offset, count, expected",
        (
            (5, 0, None, [5, 7, 9, 11]),
            (5, 0, 2, [5, 7]),
            (4, -2, 3, [1, 3, 5]),
            (0, 0, 3, [1, 3, 5]),
            (150, 0, None, []),
        ),
    )
    def test_list_remove_by_value_rank_range_relative(self, value, offset, count, expected):
        """
        Without a return type this should return the value
        """
        operation = list_operations.list_remove_by_value_rank_range_relative(
            self.test_bin, value, offset, return_type=aerospike.LIST_RETURN_VALUE, count=count
        )

        result = get_list_result_from_operation(self.as_connection, self.test_key, operation, self.test_bin)
        assert result == expected

        _, _, bins = self.as_connection.get(self.test_key)
        list_items = bins[self.test_bin]

        # Ensure that the correct number of items were removed
        assert len(list_items) == (len(self.test_list) - len(expected))

        # Ensure that the expected items were removed
        for item in list_items:
            assert item not in expected

    # TODO: duplicate test name?
    @pytest.mark.parametrize(
        "return_type, expected",
        (
            (aerospike.LIST_RETURN_VALUE, 5),
            (aerospike.LIST_RETURN_INDEX, 2),
            (aerospike.LIST_RETURN_RANK, 2),
            (aerospike.LIST_RETURN_REVERSE_RANK, 3),
            (aerospike.LIST_RETURN_REVERSE_INDEX, 3),
        ),
    )
    def test_list_remove_by_value_rank_range_relative1(self, return_type, expected):
        """
        Without a return type this should return the value
        """
        operation = list_operations.list_remove_by_value_rank_range_relative(
            self.test_bin, 5, 0, return_type=return_type, count=1
        )

        result = get_list_result_from_operation(self.as_connection, self.test_key, operation, self.test_bin)
        assert result == [expected]

    def test_list_remove_by_value_rank_range_relative_ret_none(self):
        """
        Without a return type this should return the value
        """
        operation = list_operations.list_remove_by_value_rank_range_relative(
            self.test_bin, 5, 0, return_type=aerospike.LIST_RETURN_NONE, count=1
        )

        _, _, bins = self.as_connection.operate(self.test_key, [operation])
        assert bins == {}

    def test_list_remove_by_value_rank_range_relative_ret_count(self):
        """
        Without a return type this should return the value
        """
        operation = list_operations.list_remove_by_value_rank_range_relative(
            self.test_bin, 5, 0, return_type=aerospike.LIST_RETURN_COUNT, count=1
        )

        result = get_list_result_from_operation(self.as_connection, self.test_key, operation, self.test_bin)
        assert result == 1

    @pytest.mark.parametrize(
        "value, offset, count, expected_err",
        (
            (5, "zero", None, e.ParamError),
            (5, 0, "None", e.ParamError),
            (5, 1.5, "None", e.ParamError),
            (5, 0, 1.5, e.ParamError),
        ),
    )
    def test_list_remove_by_value_rank_invalid_params(self, value, offset, count, expected_err):
        """
        Without a return type this should return the value
        """
        operation = list_operations.list_remove_by_value_rank_range_relative(
            self.test_bin, value, offset, return_type=aerospike.LIST_RETURN_VALUE, count=count
        )

        with pytest.raises(expected_err):
            self.as_connection.operate(self.test_key, [operation])

    @pytest.mark.parametrize(
        "value, offset, count, expected_err",
        (
            (5, "zero", None, e.ParamError),
            (5, 0, "None", e.ParamError),
            (5, 1.5, "None", e.ParamError),
            (5, 0, 1.5, e.ParamError),
        ),
    )
    def test_list_get_by_value_rank_invalid_params(self, value, offset, count, expected_err):
        """
        Without a return type this should return the value
        """
        operation = list_operations.list_get_by_value_rank_range_relative(
            self.test_bin, value, offset, return_type=aerospike.LIST_RETURN_VALUE, count=count
        )

        with pytest.raises(expected_err):
            self.as_connection.operate(self.test_key, [operation])

    # INVERTED TESTS
    # LIST = [3, 1, 5, 7, 9, 11]
    @pytest.mark.parametrize(
        "value, offset, count, expected",
        (
            (5, 0, None, [1, 3]),
            (5, 0, 2, [1, 3, 9, 11]),
            (4, -2, 3, [7, 9, 11]),
            (0, 0, 3, [7, 9, 11]),
            (150, 0, None, [1, 3, 5, 7, 9, 11]),
        ),
    )
    def test_list_get_by_value_rank_range_relative_inverted(self, value, offset, count, expected):
        """
        Without a return type this should return the value
        """
        operation = list_operations.list_get_by_value_rank_range_relative(
            self.test_bin, value, offset, return_type=aerospike.LIST_RETURN_VALUE, count=count, inverted=True
        )

        result = get_list_result_from_operation(self.as_connection, self.test_key, operation, self.test_bin)

        assert result == expected

    @pytest.mark.parametrize(
        "value, offset, count, expected",
        (
            (5, 0, None, [1, 3]),
            (5, 0, 2, [1, 3, 9, 11]),
            (4, -2, 3, [7, 9, 11]),
            (0, 0, 3, [7, 9, 11]),
            (150, 0, None, [1, 3, 5, 7, 9, 11]),
        ),
    )
    def test_list_remove_by_value_rank_range_relative_inverted(self, value, offset, count, expected):
        """
        Without a return type this should return the value
        """
        operation = list_operations.list_remove_by_value_rank_range_relative(
            self.test_bin, value, offset, return_type=aerospike.LIST_RETURN_VALUE, count=count, inverted=True
        )

        result = get_list_result_from_operation(self.as_connection, self.test_key, operation, self.test_bin)
        assert result == expected

        _, _, bins = self.as_connection.get(self.test_key)
        list_items = bins[self.test_bin]

        # Ensure that the correct number of items were removed
        assert len(list_items) == (len(self.test_list) - len(expected))

        # Ensure that the expected items were removed
        for item in list_items:
            assert item not in expected


class TestNewRelativeMapOperations(object):
    @pytest.fixture(autouse=True)
    def setup(self, request, as_connection):
        """
        Setup Method
        """

        skip_less_than_430(self.server_version)

        self.keys = []
        # INDEXES    0, 1, 2, 3, 4, 05
        # RINDEX     5, 4, 3, 2, 1, 0
        # RANK       2, 1, 0, 3, 4, 5
        # RRANK     -4,-5,-6,-3,-2,-1
        self.test_map = {0: 6, 5: 12, 10: 18, 15: 24}

        self.test_key = "test", "demo", "new_map_op"
        self.test_bin = "map"
        self.as_connection.put(self.test_key, {self.test_bin: self.test_map})

        map_policy = {"map_write_mode": aerospike.MAP_CREATE_ONLY, "map_order": aerospike.MAP_KEY_ORDERED}
        operations = [map_ops.map_set_policy(self.test_bin, map_policy)]
        self.as_connection.operate(self.test_key, operations)
        self.keys.append(self.test_key)

        yield

        for key in self.keys:
            try:
                self.as_connection.remove(key)
            except e.AerospikeError:
                pass

    @pytest.mark.parametrize(
        "value, offset, count, expected",
        (
            (7, 0, None, [5, 10, 15]),
            (7, 1, None, [10, 15]),
            (7, -1, 2, [0, 5]),
        ),
    )
    def test_get_by_value_rank_range_rel(self, value, offset, count, expected):
        operation = map_ops.map_get_by_value_rank_range_relative(
            self.test_bin, value, offset, return_type=aerospike.MAP_RETURN_KEY, count=count, inverted=False
        )
        res = get_list_result_from_operation(self.as_connection, self.test_key, operation, self.test_bin)

        assert res == expected

    @pytest.mark.parametrize(
        "value, offset, count, expected",
        (
            (7, 0, None, [5, 10, 15]),
            (7, 1, None, [10, 15]),
            (7, -1, 2, [0, 5]),
        ),
    )
    def test_remove_by_value_rank_range_rel(self, value, offset, count, expected):
        operation = map_ops.map_remove_by_value_rank_range_relative(
            self.test_bin, value, offset, return_type=aerospike.MAP_RETURN_KEY, count=count, inverted=False
        )
        res = get_list_result_from_operation(self.as_connection, self.test_key, operation, self.test_bin)

        assert res == expected
        test_map = self.as_connection.get(self.test_key)[2][self.test_bin]

        assert len(test_map) == (len(self.test_map) - len(expected))

        for item in expected:
            assert item not in test_map

    @pytest.mark.parametrize(
        "key, offset, count, expected",
        (
            (4, 0, None, [5, 10, 15]),
            (4, 1, None, [10, 15]),
            (4, -1, 2, [0, 5]),
        ),
    )
    def test_get_by_key_index_range_rel(self, key, offset, count, expected):
        operation = map_ops.map_get_by_key_index_range_relative(
            self.test_bin, key, offset, return_type=aerospike.MAP_RETURN_KEY, count=count, inverted=False
        )
        res = get_list_result_from_operation(self.as_connection, self.test_key, operation, self.test_bin)

        assert res == expected

    @pytest.mark.parametrize(
        "key, offset, count, expected",
        (
            (4, 0, None, [5, 10, 15]),
            (4, 1, None, [10, 15]),
            (4, -1, 2, [0, 5]),
        ),
    )
    def test_remove_by_key_index_range_rel(self, key, offset, count, expected):
        operation = map_ops.map_remove_by_key_index_range_relative(
            self.test_bin, key, offset, return_type=aerospike.MAP_RETURN_KEY, count=count, inverted=False
        )
        res = get_list_result_from_operation(self.as_connection, self.test_key, operation, self.test_bin)

        assert res == expected
        test_map = self.as_connection.get(self.test_key)[2][self.test_bin]

        assert len(test_map) == (len(self.test_map) - len(expected))

        for item in expected:
            assert item not in test_map

    @pytest.mark.parametrize(
        "return_type, expected",
        (
            (aerospike.MAP_RETURN_INDEX, [1]),
            (aerospike.MAP_RETURN_REVERSE_INDEX, [2]),
            (aerospike.MAP_RETURN_RANK, [1]),
            (aerospike.MAP_RETURN_REVERSE_RANK, [2]),
            (aerospike.MAP_RETURN_VALUE, [12]),
            (aerospike.MAP_RETURN_KEY_VALUE, [5, 12]),
            (aerospike.MAP_RETURN_COUNT, 1),
        ),
    )
    def test_list_get_by_key_index_range_relative_rettype(self, return_type, expected):
        """
        Without a return type this should return the value
        """
        operation = map_ops.map_get_by_key_index_range_relative(
            self.test_bin, 5, 0, return_type=return_type, count=1, inverted=False
        )
        res = get_list_result_from_operation(self.as_connection, self.test_key, operation, self.test_bin)

        assert res == expected

    @pytest.mark.parametrize(
        "function",
        (
            map_ops.map_get_by_value_rank_range_relative,
            map_ops.map_get_by_key_index_range_relative,
            map_ops.map_remove_by_key_index_range_relative,
            map_ops.map_remove_by_value_rank_range_relative,
        ),
    )
    @pytest.mark.parametrize(
        "key, offset, count, err",
        (
            (4, "zero", None, e.ParamError),
            (4, 1, "count", e.ParamError),
        ),
    )
    def test_remove_by_key_index_range_invalid(self, function, key, offset, count, err):
        operation = function(
            self.test_bin, key, offset, return_type=aerospike.MAP_RETURN_KEY, count=count, inverted=False
        )

        with pytest.raises(err):
            self.as_connection.operate(self.test_key, [operation])



# -*- coding: utf-8 -*-

import pytest

from aerospike import exception as e
from .test_base_class import TestBaseClass
from aerospike_helpers.expressions import (
    And,
    Eq,
    IntBin,
    ListGetByIndex,
    ListGetByIndexRange,
    ListGetByIndexRangeToEnd,
    ListGetByRank,
    ListGetByRankRange,
    ListGetByRankRangeToEnd,
    ListGetByValue,
    ListGetByValueList,
    ListGetByValueRange,
    ListGetByValueRelRankRange,
    ListGetByValueRelRankRangeToEnd,
    ListSize,
    ResultType,
)

import aerospike


@pytest.mark.xfail(TestBaseClass.temporary_xfail(), reason="xfail variable set")
@pytest.mark.usefixtures("as_connection", "connection_config")
class TestSetXDRFilter(object):
    @pytest.fixture(autouse=True)
    def setup_method(self, request, as_connection):
        try:
            dc_request = "get-config:context=xdr"
            nodes = self.as_connection.get_node_names()
            node_name = nodes[0]["node_name"]

            dc_response = self.as_connection.info_single_node(dc_request, node_name)
            self.dc = dc_response.split(",")[0].split("=")[2]
            ns_request = "get-config:context=xdr;dc=%s" % self.dc
            ns_response = self.as_connection.info_single_node(ns_request, node_name)
            self.ns = ns_response.split("namespaces=")[1].split(";")[0]
        except Exception:
            pytest.skip("Could not parse a data center or namespace, xdr may be disabled, skipping set_xdr_flags.")

    def test_set_xdr_filter_pos(self):
        response = self.as_connection.set_xdr_filter(self.dc, self.ns, (Eq(IntBin("bin1"), 6).compile()))
        assert response == "xdr-set-filter:dc=%s;namespace=%s;exp=kwGTUQKkYmluMQY=\tok\n" % (self.dc, self.ns)

    def test_set_xdr_filter_large_expression_pos(self):
        bin = "ilist_bin"
        expr = And(
            Eq(
                ListGetByValueRelRankRange(
                    None,
                    aerospike.LIST_RETURN_COUNT,
                    ListGetByIndex(None, aerospike.LIST_RETURN_VALUE, ResultType.INTEGER, 0, bin),
                    1,
                    3,
                    bin,
                ),
                2,
            ),
            Eq(
                ListGetByValue(
                    None,
                    aerospike.LIST_RETURN_INDEX,
                    6,
                    ListGetByValueRange(None, aerospike.LIST_RETURN_VALUE, 1, 7, bin),
                ),
                [2],
            ),
            Eq(
                ListGetByValueList(
                    None,
                    aerospike.LIST_RETURN_COUNT,
                    [2, 6],
                    ListGetByValueRelRankRangeToEnd(None, aerospike.LIST_RETURN_VALUE, 1, 1, bin),
                ),
                2,
            ),
            Eq(
                ListGetByIndexRangeToEnd(
                    None,
                    aerospike.LIST_RETURN_COUNT,
                    1,
                    ListGetByIndexRange(
                        None,
                        aerospike.LIST_RETURN_VALUE,
                        1,
                        3,
                        bin,
                    ),
                ),
                1,
            ),
            Eq(
                ListGetByRank(
                    None,
                    aerospike.LIST_RETURN_RANK,
                    ResultType.INTEGER,
                    1,
                    ListGetByRankRangeToEnd(None, aerospike.LIST_RETURN_VALUE, 1, bin),
                ),
                1,
            ),
            Eq(ListGetByRankRange(None, aerospike.LIST_RETURN_COUNT, 1, ListSize(None, bin), bin), 2),
        ).compile()

        response = self.as_connection.set_xdr_filter(self.dc, self.ns, expr)
        expected = (
            (
                "xdr-set-filter:dc=%s;namespace=%s;exp=lxCTAZV/AgCVGwWVfwIAkxMHAJNRBKlpbGlzdF9iaW4BA5NRBKlpbGlzdF9iaW4CkwGV\
                fwQAkxYBBpV/BAC"
                "UGQcBB5NRBKlpbGlzdF9iaW6SfpECkwGVfwIAkxcFkn6SAgaVfwQAlBsHAQGTUQSpaWxpc3RfYmluApMBlX8CAJMYBQGVfwQAlBgHAQOTU\
                QSpaWxpc3RfYmluAZMBlX8CAJMVAwGVfwQ"
                "AkxoHAZNRBKlpbGlzdF9iaW4BkwGVfwIAlBoFAZV/AgCREJNRBKlpbGlzdF9iaW6TUQSpaWxpc3RfYmluAg==\tok\n"
            )
            % (self.dc, self.ns)
        )
        assert response == expected

    def test_set_xdr_filter_none_filter_pos(self):
        response = self.as_connection.set_xdr_filter(self.dc, self.ns, None)
        assert response == "xdr-set-filter:dc=%s;namespace=%s;exp=null\tok\n" % (self.dc, self.ns)

    def test_set_xdr_filter_bad_filter_neg(self):
        with pytest.raises(e.ParamError):
            self.as_connection.set_xdr_filter(self.dc, self.ns, "bad_filter")



# -*- coding: utf-8 -*-
import pytest

from .test_base_class import TestBaseClass
from aerospike import exception as e

import aerospike


class TestLog(object):
    def teardown_class(cls):
        """
        Set the class level logger to a no-op to ensure no problems later
        """
        aerospike.set_log_level(aerospike.LOG_LEVEL_ERROR)
        aerospike.set_log_handler(None)

    def test_set_log_level_correct(self):
        """
        Test log level with correct parameters
        """

        response = aerospike.set_log_level(aerospike.LOG_LEVEL_DEBUG)

        assert response == 0

    def test_enable_log_handler_correct(self):
        """
        Test log handler with correct parameters
        """

        response = aerospike.set_log_level(aerospike.LOG_LEVEL_DEBUG)
        aerospike.set_log_handler(None)

        # Forces an event to be logged
        client = TestBaseClass.get_new_connection()

        assert response == 0
        client.close()

    @pytest.mark.parametrize("level", [None, [], {}, 1.5, "serious"])
    def test_set_log_level_with_invalid_type(self, level):
        """
        Test set_log_level with non int subtypes
        """
        with pytest.raises(e.ParamError) as param_error:
            aerospike.set_log_level(level)

        assert param_error.value.code == -2

    @pytest.mark.skip(reason="This behavior may or may not be correct")
    def test_set_log_level_with_bool(self):
        """
        Test log level with log level as a bool,
        this works because bool is a subclass of int
        """
        with pytest.raises(e.ParamError) as param_error:
            aerospike.set_log_level(False)

        assert param_error.value.code == -2

    def test_set_log_level_incorrect(self):
        """
        Test log level with a log level of valid type, but outside of
        expected range
        """
        response = aerospike.set_log_level(9)

        assert response == 0



# -*- coding: utf-8 -*-
##########################################################################
# Copyright 2013-2021 Aerospike, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
##########################################################################

from __future__ import print_function

import aerospike
import sys
import random
import time
import threading

from optparse import OptionParser
from aerospike import exception as e

##########################################################################
# Options Parsing
##########################################################################

usage = "usage: %prog [options]"

optparser = OptionParser(usage=usage, add_help_option=False)

optparser.add_option(
    "--help", dest="help", action="store_true",
    help="Displays this message.")

optparser.add_option(
    "-U", "--username", dest="username", type="string", metavar="<USERNAME>",
    help="Username to connect to database.")

optparser.add_option(
    "-P", "--password", dest="password", type="string", metavar="<PASSWORD>",
    help="Password to connect to database.")

optparser.add_option(
    "-h", "--host", dest="host", type="string", default="127.0.0.1",
    metavar="<ADDRESS>",
    help="Address of Aerospike server.")

optparser.add_option(
    "-p", "--port", dest="port", type="int", default=3000, metavar="<PORT>",
    help="Port of the Aerospike server.")

(options, args) = optparser.parse_args()

if options.help:
    optparser.print_help()
    print()
    sys.exit(1)

##########################################################################
# Client Configuration
##########################################################################

config = {
    'hosts': [(options.host, options.port)],
    'lua': {'user_path': '.'}
}

##########################################################################
# Application
##########################################################################

try:
    client = aerospike.client(config).connect(
        options.username, options.password)
except e.ClientError as exception:
    print('Error: {0} [{1}]'.format(exception.msg, exception.code))
    sys.exit(1)

namespace = 'test'
testSet = 'test'
numKeys = 10000
numReads = 1000000
fNames = ('Jimmy', 'Johnny', 'Sammy', 'Sally', 'Sandy', 'Mandy', 'Billy')
lNames = ('Bama', 'Mama', 'Sama', 'Lama', 'Cama', 'Rama', 'Tama')
numThreads = 5


def writeWork(nKeys):
    t0 = float(time.time())

    for x in range(0, nKeys):
        kstr = 'k' + str(x)
        key = (namespace, testSet, kstr)

        try:
            # Write a record
            client.put(key, {
                'name': random.choice(fNames) + ' ' + random.choice(lNames),
                'age': random.randint(10, 100),
                'value': x
            })
        except Exception as e:
            print('write error {0}'.format(e))

        if x % 1000 == 0 and x > 0:
            print('Wrote {0} records at T = {1:.2f} sec'.format(
                                     x, float(time.time()) - t0))

    print('Wrote {0} records at T = {1:.2f} sec'.format(
                        nKeys, float(time.time()) - t0))


def readWork(nReads, thrName):
    print('Thread #{0} is starting to read {1} records'.format(
                                thrName, nReads))

    # Read records
    t0 = float(time.time())

    for x in range(0, nReads):
        kstr = 'k' + str(random.randrange(0, numKeys))
        key = (namespace, testSet, kstr)
        try:
            (key, _, _) = client.get(key)
        except aerospike.exception.ClientError as e:
            print('Aerospike Error: {0} [{1}]'.format(e.msg, e.code))
            return None

        if x % 10000 == 0 and x > 0:
            print('Thread #{0} : Read {1} records at T = {2:.2f} sec'.format(
                        thrName, x, float(time.time()) - t0))
    print('Thread #{0} : Read {1} records at T = {2:.2f} sec'.format(
                        thrName, nReads, float(time.time()) - t0))


print('Writing data into Aerospike DB')
writeWork(numKeys)

print('Reading data from Aerospike DB using {0} threads'.format(numThreads))
t = []

for i in range(numThreads):
    thread = threading.Thread(target=readWork,
                              args=(numReads // numThreads, str(i)))
    thread.start()
    t.append(thread)

for i in range(numThreads):
    t[i].join()

print('Finished. Closing Aerospike connection.')
client.close()



# -*- coding: utf-8 -*-
import pytest
import time

from .test_base_class import TestBaseClass
from aerospike import exception as e

import aerospike


@pytest.mark.xfail(TestBaseClass.temporary_xfail(), reason="xfail variable set")
@pytest.mark.usefixtures("as_connection", "connection_config")
class TestInfoSingleNode(object):
    pytest_skip = False

    @pytest.fixture(autouse=True)
    def setup(self, request, as_connection, connection_config):
        key = ("test", "demo", "list_key")
        rec = {"names": ["John", "Marlen", "Steve"]}
        self.host_name = list(self.as_connection.info("node").keys())[0]
        self.as_connection.put(key, rec)

        yield

        self.as_connection.remove(key)

    def test_info_single_node_positive(self):
        """
        Test info with correct arguments.
        """
        if self.pytest_skip:
            pytest.xfail()
        response = self.as_connection.info_single_node("bins", self.host_name)

        # This should probably make sure that a bin is actually named 'names'
        assert "names" in response

    def test_info_single_node_positive_for_namespace(self):
        """
        Test info with 'namespaces' as the command.
        """
        if self.pytest_skip:
            pytest.xfail()
        response = self.as_connection.info_single_node("namespaces", self.host_name)

        assert "test" in response

    def test_info_single_node_positive_for_sets(self):
        """
        Test info with 'sets' as the command.
        """
        if self.pytest_skip:
            pytest.xfail()
        response = self.as_connection.info_single_node("sets", self.host_name)

        assert "demo" in response

    def test_info_single_node_positive_for_sindex_creation(self):
        """
        Test creating an index through an info call.
        """
        if self.pytest_skip:
            pytest.xfail()
        try:
            self.as_connection.index_remove("test", "names_test_index")
            time.sleep(2)
        except Exception:
            pass

        self.as_connection.info_single_node(
            "sindex-create:ns=test;set=demo;indexname=names_test_index;indexdata=names,string", self.host_name
        )
        time.sleep(2)

        response = self.as_connection.info_single_node("sindex", self.host_name)
        self.as_connection.index_remove("test", "names_test_index")
        assert "names_test_index" in response

    def test_info_single_node_positive_with_correct_policy(self):
        """
        Test info call with bins as command and a timeout policy.
        """
        if self.pytest_skip:
            pytest.xfail()
        host = self.host_name
        policy = {"timeout": 1000}
        response = self.as_connection.info_single_node("bins", host, policy)

        assert "names" in response

    def test_info_single_node_positive_with_host(self):
        """
        Test info with correct host.
        """
        if self.pytest_skip:
            pytest.xfail()
        host = self.host_name
        response = self.as_connection.info_single_node("bins", host)

        assert "names" in response

    def test_info_single_node_positive_with_all_parameters(self):
        """
        Test info with all parameters.
        """
        if self.pytest_skip:
            pytest.xfail()
        policy = {"timeout": 1000}
        host = self.host_name
        response = self.as_connection.info_single_node("logs", host, policy)

        assert response is not None


# Tests for incorrect usage
@pytest.mark.xfail(TestBaseClass.temporary_xfail(), reason="xfail variable set")
@pytest.mark.usefixtures("as_connection", "connection_config")
class TestInfoSingleNodeIncorrectUsage(object):
    """
    Tests for invalid usage of the the info_single_node method.
    """

    def test_info_single_node_no_parameters(self):
        """
        Test info with no parameters.
        """
        with pytest.raises(TypeError) as typeError:
            self.as_connection.info_single_node()
        assert "argument 'command' (pos 1)" in str(typeError.value)

    def test_info_single_node_for_incorrect_command(self):
        """
        Test info for incorrect command.
        """
        with pytest.raises(e.ClientError):
            self.as_connection.info_single_node("abcd", self.connection_config["hosts"][0])

    def test_info_single_node_positive_without_connection(self):
        """
        Test info with correct arguments without connection.
        """
        client1 = aerospike.client(self.connection_config)
        client1.close()
        with pytest.raises(e.ClusterError) as err_info:
            client1.info_single_node("bins", self.connection_config["hosts"][0][:2])

        assert err_info.value.code == 11
        assert err_info.value.msg == "No connection to aerospike cluster."

    def test_info_single_node_positive_with_extra_parameters(self):
        """
        Test info with extra parameters.
        """
        host = self.connection_config["hosts"][0]
        policy = {"timeout": 1000}
        with pytest.raises(TypeError) as typeError:
            self.as_connection.info_single_node("bins", host, policy, "")

        assert "info_single_node() takes at most 3 arguments (4 given)" in str(typeError.value)

    def test_info_single_node_positive_with_incorrect_host(self):
        """
        Test info with incorrect host.
        """
        host = "wrong"
        with pytest.raises(e.ParamError):
            self.as_connection.info_single_node("bins", host)

    @pytest.mark.parametrize("command", (None, 5, ["info"], {}, False))
    def test_info_single_node_for_invalid_command_type(self, command):
        """
        Test info for None command.
        """
        with pytest.raises(e.ParamError):
            self.as_connection.info_single_node(command, self.connection_config["hosts"][0][:2])

    @pytest.mark.parametrize("hostname", (None, 5, ["localhost"], {}, 3000.0))
    def test_info_single_node_for_invalid_hostname_type(self, hostname):
        """
        Test info for invalid hostname types.
        """
        with pytest.raises(e.ParamError):
            self.as_connection.info_single_node("info", hostname)

    def test_info_single_node_positive_with_incorrect_policy(self):
        """
        Test info with incorrect policy.
        """
        host = "host"
        policy = {"timeout": 0.5}
        with pytest.raises(e.ParamError) as err_info:
            self.as_connection.info_single_node("bins", host, policy)

        assert err_info.value.code == -2
        assert err_info.value.msg == "timeout is invalid"

    @pytest.mark.parametrize("host", ([(3000, 3000)], [], "123.456:1000", 3000, None))
    def test_info_single_node_positive_with_incorrect_host_type(self, host):
        """
        Test info with incorrect host type.
        """

        with pytest.raises(e.ParamError):
            self.as_connection.info_single_node("bins", host)



# -*- coding: utf-8 -*-
import pytest

from .test_base_class import TestBaseClass
import aerospike
from aerospike import exception as e


class TestPrepend:
    @pytest.fixture(autouse=True)
    def setup(self, request, as_connection):
        """
        Setup Method
        """
        for i in range(5):
            key = ("test", "demo", i)
            rec = {"name": "name%s" % (str(i)), "age": i, "nolist": [1, 2, 3]}
            as_connection.put(key, rec)

        key = ("test", "demo", "bytearray_key")
        as_connection.put(key, {"bytearray_bin": bytearray("asd;as[d'as;d", "utf-8")})

        key = ("test", "demo", "bytes_key")
        as_connection.put(key, {"bytes_bin": b"x"})

        def teardown():
            """
            Teardown Method
            """
            for i in range(5):
                key = ("test", "demo", i)
                as_connection.remove(key)

            key = ("test", "demo", "bytearray_key")
            as_connection.remove(key)

            key = ("test", "demo", "bytes_key")
            as_connection.remove(key)

        request.addfinalizer(teardown)

    def test_pos_prepend_with_correct_parameters(self):
        """
        Invoke prepend() with correct parameters
        """
        key = ("test", "demo", 1)
        self.as_connection.prepend(key, "name", "str")

        (key, _, bins) = self.as_connection.get(key)

        assert bins == {"age": 1, "name": "strname1", "nolist": [1, 2, 3]}

    def test_pos_prepend_with_correct_policy(self):
        """
        Invoke prepend() with correct policy
        """
        key = ("test", "demo", 1)
        policy = {"timeout": 1000, "key": aerospike.POLICY_KEY_SEND, "commit_level": aerospike.POLICY_COMMIT_LEVEL_ALL}

        self.as_connection.prepend(key, "name", "str", {}, policy)

        (key, _, bins) = self.as_connection.get(key)

        assert bins == {"age": 1, "name": "strname1", "nolist": [1, 2, 3]}

    def test_pos_prepend_with_policy_key_send(self):
        """
        Invoke prepend() with policy key send
        """
        key = ("test", "demo", 1)
        policy = {
            "timeout": 1000,
            "key": aerospike.POLICY_KEY_SEND,
            "retry": aerospike.POLICY_RETRY_ONCE,
            "commit_level": aerospike.POLICY_COMMIT_LEVEL_MASTER,
        }
        self.as_connection.prepend(key, "name", "str", {}, policy)

        (key, _, bins) = self.as_connection.get(key)

        assert bins == {"age": 1, "name": "strname1", "nolist": [1, 2, 3]}
        assert key == (
            "test",
            "demo",
            None,
            bytearray(b"\xb7\xf4\xb88\x89\xe2\xdag\xdeh>\x1d\xf6\x91\x9a\x1e\xac\xc4F\xc8"),
        )

    def test_pos_prepend_with_policy_key_gen_EQ_ignore(self):
        """
        Invoke prepend() with gen eq positive
        """
        key = ("test", "demo", 1)
        policy = {
            "timeout": 1000,
            "key": aerospike.POLICY_KEY_SEND,
            "retry": aerospike.POLICY_RETRY_ONCE,
            "gen": aerospike.POLICY_GEN_IGNORE,
        }

        meta = {"gen": 10, "ttl": 1200}
        self.as_connection.prepend(key, "name", "str", meta, policy)

        (key, meta, bins) = self.as_connection.get(key)

        assert bins == {"age": 1, "name": "strname1", "nolist": [1, 2, 3]}
        assert key == (
            "test",
            "demo",
            None,
            bytearray(b"\xb7\xf4\xb88\x89\xe2\xdag\xdeh>\x1d\xf6\x91\x9a\x1e\xac\xc4F\xc8"),
        )

    def test_pos_prepend_with_policy_key_gen_EQ_positive(self):
        """
        Invoke prepend() with gen eq positive
        """
        key = ("test", "demo", 1)
        policy = {
            "timeout": 1000,
            "key": aerospike.POLICY_KEY_SEND,
            "retry": aerospike.POLICY_RETRY_ONCE,
            "gen": aerospike.POLICY_GEN_EQ,
        }
        (key, meta) = self.as_connection.exists(key)

        gen = meta["gen"]
        meta = {"gen": gen, "ttl": 1200}
        self.as_connection.prepend(key, "name", "str", meta, policy)

        (key, meta, bins) = self.as_connection.get(key)

        assert bins == {"age": 1, "name": "strname1", "nolist": [1, 2, 3]}
        assert key == (
            "test",
            "demo",
            None,
            bytearray(b"\xb7\xf4\xb88\x89\xe2\xdag\xdeh>\x1d\xf6\x91\x9a\x1e\xac\xc4F\xc8"),
        )

    def test_pos_prepend_with_policy_key_gen_GT_positive(self):
        """
        Invoke prepend() with gen GT positive
        """
        key = ("test", "demo", 1)
        policy = {
            "timeout": 1000,
            "key": aerospike.POLICY_KEY_SEND,
            "retry": aerospike.POLICY_RETRY_ONCE,
            "gen": aerospike.POLICY_GEN_GT,
        }
        (key, meta) = self.as_connection.exists(key)

        gen = meta["gen"]
        meta = {"gen": gen + 2, "ttl": 1200}
        self.as_connection.prepend(key, "name", "str", meta, policy)

        (key, meta, bins) = self.as_connection.get(key)

        assert bins == {"age": 1, "name": "strname1", "nolist": [1, 2, 3]}
        assert key == (
            "test",
            "demo",
            None,
            bytearray(b"\xb7\xf4\xb88\x89\xe2\xdag\xdeh>\x1d\xf6\x91\x9a\x1e\xac\xc4F\xc8"),
        )

    def test_pos_prepend_with_policy_key_digest(self):
        """
        Invoke prepend() with policy key digest
        """
        key = ("test", "demo", None, bytearray("asd;as[d'as;djk;uyfl", "utf-8"))
        rec = {"name": "name%s" % (str(1)), "age": 1, "nolist": [1, 2, 3]}
        self.as_connection.put(key, rec)

        policy = {"timeout": 1000, "key": aerospike.POLICY_KEY_DIGEST, "retry": aerospike.POLICY_RETRY_NONE}
        self.as_connection.prepend(key, "name", "str", {}, policy)

        (key, _, bins) = self.as_connection.get(key)

        assert bins == {"age": 1, "name": "strname1", "nolist": [1, 2, 3]}
        assert key == ("test", "demo", None, bytearray(b"asd;as[d'as;djk;uyfl"))

        self.as_connection.remove(key)

    @pytest.mark.parametrize(
        "key, bin, value, expected",
        [(("test", "demo", 1), "name", "age", "agename1"), (("test", "demo", 1), "add", "address", "address")],
    )
    def test_pos_prepend_unicode_parameters(self, key, bin, value, expected):
        """
        Invoke prepend() with unicode parameters
        """
        self.as_connection.prepend(key, bin, value)

        key, _, bins = self.as_connection.get(key)
        assert bins[bin] == expected

    def test_pos_prepend_key_with_none_set_name(self):
        """
        Invoke get_many with none set name
        """
        key = ("test", None, 1)
        policy = {"timeout": 1000}
        self.as_connection.prepend(key, "name", "ABC", {}, policy)
        key, _, bins = self.as_connection.get(key)
        assert bins == {"name": "ABC"}
        self.as_connection.remove(key)

    def test_pos_prepend_with_nonexistent_key(self):
        """
        Invoke prepend() with non-existent key
        """
        key = ("test", "demo", 1000)
        status = self.as_connection.prepend(key, "name", "str")

        assert status == 0
        key, _, bins = self.as_connection.get(key)
        assert bins["name"] == "str"

        self.as_connection.remove(key)

    def test_pos_prepend_with_nonexistent_bin(self):
        """
        Invoke prepend() with non-existent bin
        """
        key = ("test", "demo", 1)
        status = self.as_connection.prepend(key, "name1", "str")

        assert status == 0
        key, _, bins = self.as_connection.get(key)
        assert bins["name1"] == "str"

    def test_pos_prepend_with_bytearray(self):
        """
        Invoke prepend() with bytearray value
        """
        key = ("test", "demo", "bytearray_key")
        self.as_connection.prepend(key, "bytearray_bin", bytearray("abc", "utf-8"))

        (key, _, bins) = self.as_connection.get(key)

        assert bins == {"bytearray_bin": bytearray("abcasd;as[d'as;d", "utf-8")}

    def test_pos_prepend_with_bytearray_new_key(self):
        """
        Invoke prepend() with bytearray value with a new record(non-existing)
        """
        key = ("test", "demo", "bytearray_new")
        self.as_connection.prepend(key, "bytearray_bin", bytearray("asd;as[d'as;d", "utf-8"))

        (key, _, bins) = self.as_connection.get(key)

        assert bins == {"bytearray_bin": bytearray("asd;as[d'as;d", "utf-8")}

        self.as_connection.remove(key)

    def test_pos_prepend_with_bytes(self):
        """
        Invoke prepend() with bytes value
        """
        key = ("test", "demo", "bytes_key")
        self.as_connection.prepend(key, "bytes_bin", b"a")

        (key, _, bins) = self.as_connection.get(key)

        assert bins == {"bytes_bin": b"ax"}

    def test_pos_prepend_with_bytes_new_key(self):
        """
        Invoke prepend() with bytes value with a new record(non-existing)
        """
        key = ("test", "demo", "bytes_new")
        self.as_connection.prepend(key, "bytes_bin", b"a")

        (key, _, bins) = self.as_connection.get(key)

        assert bins == {"bytes_bin": b"a"}

        self.as_connection.remove(key)

    # Negative tests
    def test_neg_prepend_with_no_parameters(self):
        """
        Invoke prepend() without any mandatory parameters.
        """
        with pytest.raises(TypeError) as typeError:
            self.as_connection.prepend()
        assert "argument 'key' (pos 1)" in str(typeError.value)

    def test_neg_prepend_with_policy_key_gen_EQ_not_equal(self):
        """
        Invoke prepend() with policy key GEN_EQ not equal
        """
        key = ("test", "demo", 1)
        policy = {
            "timeout": 1000,
            "key": aerospike.POLICY_KEY_SEND,
            "retry": aerospike.POLICY_RETRY_ONCE,
            "gen": aerospike.POLICY_GEN_EQ,
        }
        (key, meta) = self.as_connection.exists(key)
        gen = meta["gen"]

        meta = {"gen": gen + 5, "ttl": 1200}
        try:
            self.as_connection.prepend(key, "name", "str", meta, policy)

        except e.RecordGenerationError as exception:
            assert exception.code == 3
            assert exception.bin == "name"

        (key, meta, bins) = self.as_connection.get(key)

        assert bins == {"age": 1, "name": "name1", "nolist": [1, 2, 3]}
        assert key == (
            "test",
            "demo",
            None,
            bytearray(b"\xb7\xf4\xb88\x89\xe2\xdag\xdeh>\x1d\xf6\x91\x9a\x1e\xac\xc4F\xc8"),
        )

    def test_neg_prepend_with_policy_key_gen_GT_lesser(self):
        """
        Invoke prepend() with gen GT positive lesser
        """
        key = ("test", "demo", 1)
        policy = {
            "timeout": 1000,
            "key": aerospike.POLICY_KEY_SEND,
            "retry": aerospike.POLICY_RETRY_ONCE,
            "gen": aerospike.POLICY_GEN_GT,
        }
        (key, meta) = self.as_connection.exists(key)

        gen = meta["gen"]
        meta = {"gen": gen, "ttl": 1200}
        try:
            self.as_connection.prepend(key, "name", "str", meta, policy)

        except e.RecordGenerationError as exception:
            assert exception.code == 3
            assert exception.bin == "name"

        (key, meta, bins) = self.as_connection.get(key)

        assert bins == {"age": 1, "name": "name1", "nolist": [1, 2, 3]}
        assert key == (
            "test",
            "demo",
            None,
            bytearray(b"\xb7\xf4\xb88\x89\xe2\xdag\xdeh>\x1d\xf6\x91\x9a\x1e\xac\xc4F\xc8"),
        )

    def test_neg_prepend_with_incorrect_policy(self):
        """
        Invoke prepend() with incorrect policy
        """
        key = ("test", "demo", 1)
        policy = {"timeout": 0.5}
        try:
            self.as_connection.prepend(key, "name", "str", {}, policy)

        except e.ParamError as exception:
            assert exception.code == -2
            assert exception.msg == "timeout is invalid"

    @pytest.mark.parametrize(
        "key, bin, value, meta, policy, ex_code, ex_msg",
        [
            (("test", "demo", 1), "name", 2, {}, {}, -2, "Cannot concatenate 'str' and 'non-str' objects"),
            (("test", "demo", 1), "name", "abc", {}, "", -2, "policy must be a dict"),
            (("test", "demo", 1), 1, "ABC", {}, {"timeout": 1000}, -2, "Bin name should be of type string"),
        ],
    )
    def test_neg_prepend_parameters_incorrect_datatypes(self, key, bin, value, meta, policy, ex_code, ex_msg):
        """
        Invoke prepend() with parameters of incorrect datatypes
        """
        try:
            self.as_connection.prepend(key, bin, value, meta, policy)

        except e.ParamError as exception:
            assert exception.code == ex_code
            assert exception.msg == ex_msg

    def test_neg_prepend_with_extra_parameter(self):
        """
        Invoke prepend() with extra parameter.
        """
        key = ("test", "demo", 1)
        policy = {"timeout": 1000}
        with pytest.raises(TypeError) as typeError:
            self.as_connection.prepend(key, "name", "str", {}, policy, "")

        assert "prepend() takes at most 5 arguments (6 given)" in str(typeError.value)

    @pytest.mark.parametrize(
        "key, bin, value, ex_code, ex_msg",
        [
            (None, "name", "str", -2, "key is invalid"),
            (("test", "demo", 1), None, "str", -2, "Bin name should be of type string"),
        ],
    )
    def test_neg_prepend_parameters_as_none(self, key, bin, value, ex_code, ex_msg):
        """
        Invoke prepend() with parameters as None
        """
        try:
            self.as_connection.prepend(key, bin, value)

        except e.ParamError as exception:
            assert exception.code == ex_code
            assert exception.msg == ex_msg

    def test_neg_prepend_invalid_key_invalid_ns(self):
        """
        Invoke prepend() invalid namespace
        """
        key = ("test1", "demo", 1)
        policy = {"timeout": 1000}
        with pytest.raises(e.ClientError):
            self.as_connection.prepend(key, "name", "ABC", {}, policy)

    @pytest.mark.parametrize(
        "key, bin, value, meta, policy, ex_code, ex_msg",
        [
            (
                ("test", 1),
                "name",
                "ABC",
                {},
                {"timeout": 1000},
                -2,
                "key tuple must be (Namespace, Set, Key) or (Namespace, Set, None, Digest)",
            ),
            (
                ("test", "set"),
                "name",
                "ABC",
                {},
                {"timeout": 1000},
                -2,
                "key tuple must be (Namespace, Set, Key) or (Namespace, Set, None, Digest)",
            ),
            (
                ("test", "demo", 1, 1, 1),
                "name",
                "ABC",
                {},
                {"timeout": 1000},
                -2,
                "key tuple must be (Namespace, Set, Key) or (Namespace, Set, None, Digest)",
            ),
        ],
    )
    def test_neg_prepend_invalid_key_combinations(self, key, bin, value, meta, policy, ex_code, ex_msg):
        """
        Invoke prepend() with invalid key combinations
        """
        try:
            self.as_connection.prepend(key, bin, value, meta, policy)
            key, meta, _ = self.as_connection.get(key)

        except e.ParamError as exception:
            assert exception.code == ex_code
            assert exception.msg == ex_msg

    def test_neg_prepend_without_bin_name(self):
        """
        Invoke prepend without bin name
        """
        key = ("test", "demo", 1)
        policy = {"timeout": 1000}
        try:
            self.as_connection.prepend(key, "ABC", {}, policy)
            key, _, _ = self.as_connection.get(key)
        except e.ParamError as exception:
            assert exception.code == -2
            assert exception.msg == "Cannot concatenate 'str' and 'non-str' objects"

    def test_neg_prepend_with_correct_parameters_without_connection(self):
        """
        Invoke prepend() with correct parameters without connection
        """
        config = TestBaseClass.get_connection_config()
        client1 = aerospike.client(config)
        client1.close()
        key = ("test", "demo", 1)

        try:
            client1.prepend(key, "name", "str")

        except e.ClusterError as exception:
            assert exception.code == 11
